"""Mysql module tests."""

import json
import logging
import re
from datetime import datetime
from decimal import Decimal
from unittest import mock

import pytest
from ClusterShell.MsgTree import MsgTreeElem
from cumin import Config, nodeset
from cumin.transports import Command
from pymysql.cursors import DictCursor

from spicerack import mysql
from spicerack.constants import WMF_CA_BUNDLE_PATH
from spicerack.remote import Remote, RemoteExecutionError, RemoteHosts
from spicerack.tests import get_fixture_path
from spicerack.tests.unit.test_remote import mock_cumin

EQIAD_CORE_MASTERS_QUERY = "db10[01-11]"
CODFW_CORE_MASTERS_QUERY = "db20[01-11]"
VERTICAL_QUERY_NEWLINE = """*************************** 1. row ***************************
test: line with
a newline
"""
MASTER_STATUS = {"Binlog_Do_DB": "", "Binlog_Ignore_DB": "", "File": "host1-bin.001234", "Position": 123456782}


class TestInstance:
    """Test class for the Instance class."""

    @mock.patch("spicerack.mysql.MysqlClient", autospec=True)
    def setup_method(self, _, mocked_pymysql_connection):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_pymysql = mocked_pymysql_connection.return_value
        self.mocked_cursor = (
            self.mocked_pymysql.connect.return_value.__enter__.return_value.cursor.return_value.__enter__.return_value
        )
        self.mocked_run_sync = mock.Mock()
        self.config = Config(get_fixture_path("remote", "config.yaml"))
        single = RemoteHosts(self.config, nodeset("single1"), dry_run=False)
        single_dry_run = RemoteHosts(self.config, nodeset("single1"))
        single.run_sync = self.mocked_run_sync
        multi = RemoteHosts(self.config, nodeset("multi1"), dry_run=False)
        multi.run_sync = self.mocked_run_sync
        self.single_instance = mysql.Instance(single)
        self.multi_instance = mysql.Instance(multi, name="instance1")
        self.single_instance_dry_run = mysql.Instance(single_dry_run)

    def test_init_raise(self):
        """It should raise a NotImplementedError exception if more than one host is passed to the constructor."""
        with pytest.raises(NotImplementedError, match="Only single hosts are currently supported"):
            mysql.Instance(RemoteHosts(self.config, nodeset("host[1-2]")))

    @pytest.mark.parametrize(
        "instance, expected",
        (("single_instance", "single1 (single-instance)"), ("multi_instance", "multi1 (instance1)")),
    )
    def test_str(self, instance, expected):
        """It should return a string representation of the instance."""
        assert str(getattr(self, instance)) == expected

    @pytest.mark.parametrize(
        "instance, expected",
        [
            ("single_instance", "/srv/sqldata"),
            ("multi_instance", "/srv/sqldata.instance1"),
        ],
    )
    def test_data_dir(self, instance, expected):
        """It should return the correct data directory path for single and multi instances."""
        assert getattr(self, instance).data_dir == expected

    @pytest.mark.parametrize(
        "instance, expected",
        [
            ("single_instance", "/run/mysqld/mysqld.sock"),
            ("multi_instance", "/run/mysqld/mysqld.instance1.sock"),
        ],
    )
    def test_socket(self, instance, expected):
        """It should return the correct socket path for single and multi instances."""
        assert getattr(self, instance).socket == expected

    def test_cursor(self):
        """It should allow to perform queries on the target instance via the pymysql library."""
        with self.single_instance.cursor(database="mydatabase") as (connection, cursor):
            assert connection is self.mocked_pymysql.connect.return_value.__enter__.return_value
            assert self.mocked_pymysql.connect.call_args.kwargs["database"] == "mydatabase"
            cursor.execute("SELECT * FROM table")
            cursor.fetchall()

        self.mocked_cursor.execute.assert_called_once_with("SELECT * FROM table")
        self.mocked_cursor.fetchall.assert_called_once_with()

    def test_check_warnings_absent(self):
        """It should just return if there are no warnings."""
        self.mocked_cursor.execute.return_value = 0
        with self.single_instance.cursor() as (_connection, cursor):
            cursor.execute("SELECT 1")
            self.single_instance.check_warnings(cursor)

        self.mocked_cursor.execute.assert_called_with("SHOW WARNINGS")

    @mock.patch("spicerack.mysql.ask_confirmation")
    def test_check_warnings_present(self, mocked_ask_confirmation, caplog):
        """It should log the warnings and ask the operator what to do if there are warnings."""
        self.mocked_cursor.execute.side_effect = [0, 1]
        self.mocked_cursor.fetchall.return_value = [{"Level": "Warning", "Code": 123, "Message": "Some error"}]

        with caplog.at_level(logging.WARNING):
            with self.single_instance.cursor() as (_connection, cursor):
                cursor.execute("SELECT 1")
                self.single_instance.check_warnings(cursor)

        self.mocked_cursor.execute.assert_called_with("SHOW WARNINGS")
        mocked_ask_confirmation.assert_called_once_with(
            "The above warnings were raised during the last query, do you want to proceed anyway?"
        )
        assert "[Warning] 123: Some error" in caplog.text

    @pytest.mark.parametrize(
        "instance, is_safe",
        (
            ("single_instance", False),
            ("single_instance", True),
            ("single_instance_dry_run", True),
        ),
    )
    def test_execute(self, instance, is_safe):
        """It should execute a query within a cursor context just returning the number of affected rows."""
        self.mocked_cursor.execute.side_effect = [2, 0]
        num_rows = getattr(self, instance).execute("RESET SLAVE ALL", is_safe=is_safe)
        assert num_rows == 2
        self.mocked_cursor.execute.assert_has_calls([mock.call("RESET SLAVE ALL", None), mock.call("SHOW WARNINGS")])

    def test_execute_dry_run_unsafe(self):
        """It should not execute the query and return 0."""
        self.mocked_cursor.mogrify.return_value = "RESET SLAVE ALL"
        num_rows = self.single_instance_dry_run.execute("RESET SLAVE ALL")
        assert num_rows == 0
        self.mocked_cursor.mogrify.assert_called_once_with("RESET SLAVE ALL", None)
        self.mocked_cursor.execute.assert_not_called()

    def test_fetch_one_row_ok(self):
        """It should return the row, checking for warnings."""
        self.mocked_cursor.execute.side_effect = [1, 0]
        self.mocked_cursor.fetchone.return_value = {"value": 1}
        row = self.single_instance.fetch_one_row("SELECT 1 AS value")
        assert row == {"value": 1}
        self.mocked_cursor.execute.assert_has_calls([mock.call("SELECT 1 AS value", None), mock.call("SHOW WARNINGS")])
        self.mocked_cursor.fetchone.assert_called_once_with()

    def test_fetch_one_row_empty(self):
        """It should return None if no rows are returned."""
        self.mocked_cursor.execute.return_value = 0
        row = self.single_instance.fetch_one_row("SELECT 1 WHERE 2 > 1")
        assert row == {}
        self.mocked_cursor.execute.assert_has_calls(
            [mock.call("SELECT 1 WHERE 2 > 1", None), mock.call("SHOW WARNINGS")]
        )
        self.mocked_cursor.fetchone.assert_not_called()

    def test_fetch_one_row_too_many(self):
        """It should raise a MysqlError if more than one row is returned."""
        self.mocked_cursor.execute.side_effect = [3, 0]
        with pytest.raises(mysql.MysqlError, match="Expected query to return zero or one row, got 3 instead"):
            self.single_instance.fetch_one_row("SELECT * FROM mytable", database="mydb")
        self.mocked_cursor.execute.assert_has_calls(
            [mock.call("SELECT * FROM mytable", None), mock.call("SHOW WARNINGS")]
        )
        self.mocked_cursor.fetchone.assert_not_called()

    @pytest.mark.parametrize(
        "query, database, kwargs",
        (
            ("SELECT 1 AS test", "", {}),
            ("SELECT 1 AS test FROM dbname", "dbname", {}),
            ("SELECT 1 AS test", "", {"is_safe": True}),
            ("SELECT 1 AS test", "", {"print_progress_bars": True, "print_output": True}),
        ),
    )
    def test_single_run_query_ok(self, query, database, kwargs):
        """It should run the query and return the results."""
        self.mocked_run_sync.return_value = [(nodeset("single1"), MsgTreeElem(b"test\n1", parent=MsgTreeElem()))]
        results = self.single_instance.run_query(query, database=database, **kwargs)
        assert [(str(host), msg.message().decode()) for host, msg in results] == [("single1", "test\n1")]
        expected_kwargs = {"print_progress_bars": False, "print_output": False, **kwargs}
        self.mocked_run_sync.assert_called_once_with(
            f'/usr/local/bin/mysql --socket /run/mysqld/mysqld.sock --batch --execute "{query}" {database}'.strip(),
            **expected_kwargs,
        )

    def test_multi_query_raise(self):
        """It should raise a MysqlError exception if the query execution fails."""
        self.mocked_run_sync.side_effect = RemoteExecutionError(retcode=1, message="error", results=iter(()))
        with pytest.raises(mysql.MysqlError, match="Failed to run 'invalid' on multi1"):
            self.multi_instance.run_query("invalid")

        self.mocked_run_sync.assert_called_once_with(
            '/usr/local/bin/mysql --socket /run/mysqld/mysqld.instance1.sock --batch --execute "invalid"',
            print_progress_bars=False,
            print_output=False,
        )

    def test_run_vertical_query_ok(self):
        """It should return the current slave status of the database."""
        status = get_fixture_path("mysql", "single_show_slave_status.out").read_text()
        self.mocked_run_sync.return_value = [
            (nodeset("single1"), MsgTreeElem("".join((status, status)).encode(), parent=MsgTreeElem()))
        ]
        rows = self.single_instance.run_vertical_query("SELECT")  # dummy query
        self.mocked_run_sync.assert_called_once_with(
            r'/usr/local/bin/mysql --socket /run/mysqld/mysqld.sock --batch --execute "SELECT\G"',
            print_progress_bars=False,
            print_output=False,
        )
        assert len(rows) == 2
        assert rows[0]["Seconds_Behind_Master"] == rows[1]["Seconds_Behind_Master"] == "0"

    def test_run_vertical_query_no_reponse(self):
        """It should return an empty list. This should not happen in real life."""
        self.mocked_run_sync.return_value = []
        rows = self.single_instance.run_vertical_query("SELECT")  # dummy query
        assert rows == []  # pylint: disable=use-implicit-booleaness-not-comparison

    def test_run_vertical_query_empty(self):
        """It should return an empty list. This should not happen in real life."""
        response = "*************************** 1. row ***************************\n".encode()
        self.mocked_run_sync.return_value = [(nodeset("single1"), MsgTreeElem(response, parent=MsgTreeElem()))]
        rows = self.single_instance.run_vertical_query("SELECT")  # dummy query
        assert rows == []  # pylint: disable=use-implicit-booleaness-not-comparison

    def test_run_vertical_query_parse_error(self, caplog):
        """It should log the skipping of some lines that were not properly parser."""
        self.mocked_run_sync.return_value = [
            (nodeset("single1"), MsgTreeElem(VERTICAL_QUERY_NEWLINE.encode(), parent=MsgTreeElem()))
        ]
        with caplog.at_level(logging.ERROR):
            rows = self.single_instance.run_vertical_query("SELECT")  # dummy query

        assert rows == [{"test": "line with"}]
        assert "Failed to parse into key/value for query 'SELECT' this line: a newline" in caplog.text

    @pytest.mark.parametrize(
        "output,expected",
        (
            (b"ActiveState=active\nSubState=running", True),
            (b"ActiveState=inactive\nSubState=running", False),
            (b"ActiveState=active\nSubState=stopped", False),
        ),
    )
    def test_is_running(self, output, expected):
        """It should return True if the service is active and running."""
        self.mocked_run_sync.return_value = [(nodeset("single1"), MsgTreeElem(output, parent=MsgTreeElem()))]
        assert self.single_instance.is_running() is expected

    @pytest.mark.parametrize("instance", ("single_instance", "multi_instance"))
    @pytest.mark.parametrize(
        "method, expected",
        (
            ("stop_slave", "STOP SLAVE"),
            ("start_slave", "START SLAVE"),
        ),
    )
    @mock.patch("spicerack.mysql.sleep", return_value=None)
    def test_run_stop_start_slave_ok(self, mocked_sleep, method, expected, instance):
        """It should run the method called and execute the related query."""
        self.mocked_cursor.execute.return_value = 0
        getattr(getattr(self, instance), method)()
        self.mocked_cursor.execute.assert_any_call(expected, None)
        assert mocked_sleep.called == (method == "start_slave")

    def test_single_show_slave_status_ok(self):
        """It should return the current slave status of the database."""
        mocked_status = json.loads(get_fixture_path("mysql", "single_show_slave_status.json").read_text())
        self.mocked_cursor.execute.return_value = 1
        self.mocked_cursor.fetchone.return_value = mocked_status

        status = self.single_instance.show_slave_status()
        assert status["Master_Host"] == "host1.example.org"
        assert status["Seconds_Behind_Master"] == 0
        assert status == mocked_status
        self.mocked_cursor.execute.assert_called_once_with("SHOW SLAVE STATUS")
        self.mocked_cursor.fetchone.assert_called_once_with()

    def test_single_show_slave_status_on_master(self):
        """It should raise a MysqlError exception if show slave status is called on a master."""
        self.mocked_cursor.execute.return_value = 0

        with pytest.raises(
            mysql.MysqlError, match=re.escape("SHOW SLAVE STATUS seems to have been executed on a master")
        ):
            self.single_instance.show_slave_status()

        self.mocked_cursor.fetchone.assert_not_called()

    def test_single_show_slave_status_multisource(self):
        """It should raise a MysqlError exception if show slave status is called on a multisource instance."""
        self.mocked_cursor.execute.return_value = 2
        with pytest.raises(NotImplementedError, match="Multisource setup are not implemented"):
            self.single_instance.show_slave_status()

        self.mocked_cursor.fetchone.assert_not_called()

    def test_single_show_master_status_ok(self):
        """It should return the current master status of the database."""
        self.mocked_cursor.execute.side_effect = [1, 0]
        self.mocked_cursor.fetchone.return_value = MASTER_STATUS

        status = self.single_instance.show_master_status()

        assert status["File"] == "host1-bin.001234"
        assert status["Position"] == 123456782
        self.mocked_cursor.execute.assert_has_calls([mock.call("SHOW MASTER STATUS", None), mock.call("SHOW WARNINGS")])
        self.mocked_cursor.fetchone.assert_called_once_with()

    def test_single_show_master_status_no_binlog(self):
        """It should raise a MysqlError if show master status is run on a host with binlog disabled."""
        self.mocked_cursor.execute.return_value = 0
        with pytest.raises(
            mysql.MysqlError,
            match=re.escape("SHOW MASTER STATUS seems to have been executed on a host with binlog disabled"),
        ):
            self.single_instance.show_master_status()

        self.mocked_cursor.execute.assert_has_calls([mock.call("SHOW MASTER STATUS", None), mock.call("SHOW WARNINGS")])
        self.mocked_cursor.fetchone.assert_not_called()

    @pytest.mark.parametrize(
        "setting, expected",
        (
            (mysql.MasterUseGTID.CURRENT_POS, "current_pos"),
            (mysql.MasterUseGTID.SLAVE_POS, "slave_pos"),
            (mysql.MasterUseGTID.NO, "no"),
        ),
    )
    def test_set_master_use_gtid_ok(self, setting, expected):
        """It should execute MASTER_USE_GTID with the given value."""
        self.mocked_cursor.execute.return_value = 0
        self.single_instance.set_master_use_gtid(setting)
        self.mocked_cursor.execute.assert_any_call(f"CHANGE MASTER TO MASTER_USE_GTID={expected}", None)

    def test_set_master_use_gtid_invalid(self):
        """It should raise MysqlError if called with an invalid setting."""
        with pytest.raises(
            mysql.MysqlError,
            match=re.escape("Only instances of MasterUseGTID are accepted, got: <class 'str'>"),
        ):
            self.single_instance.set_master_use_gtid("invalid")

    @pytest.mark.parametrize("instance", ("single_instance", "multi_instance"))
    @pytest.mark.parametrize("method", ("stop", "start", "status", "restart"))
    def test_single_systemctl_action_ok(self, method, instance):
        """It should perform the systemctl action on the mysql process."""
        self.mocked_run_sync.return_value = iter(())
        getattr(getattr(self, instance), f"{method}_mysql")()
        suffix = "@instance1" if instance == "multi_instance" else ""
        kwargs = {}
        if method == "status":
            kwargs["is_safe"] = True
        if method == "start":
            kwargs["print_output"] = True

        command = f"/usr/bin/systemctl {method} mariadb{suffix}.service"
        if method == "status":
            command = Command(command, ok_codes=[])
        self.mocked_run_sync.assert_called_once_with(command, **kwargs)

    @pytest.mark.parametrize(
        "instance, path",
        (
            ("single_instance", "/srv/sqldata"),
            ("multi_instance", "/srv/sqldata.instance1"),
        ),
    )
    @pytest.mark.parametrize("skip_confirmation", (False, True))
    @mock.patch("spicerack.mysql.ask_confirmation")
    def test_clean_data_dir(self, mocked_ask_confirmation, skip_confirmation, instance, path):
        """It should delete the data directory."""
        self.mocked_run_sync.return_value = iter(())
        obj = getattr(self, instance)
        obj.clean_data_dir(skip_confirmation=skip_confirmation)
        self.mocked_run_sync.assert_called_once_with(f"/usr/bin/rm -rf {path}")
        if skip_confirmation:
            mocked_ask_confirmation.assert_not_called()
        else:
            mocked_ask_confirmation.assert_called_once_with(
                f"ATTENTION: destructive action for {obj.host}: /usr/bin/rm -rf {path}. Are you sure to proceed?"
            )

    @pytest.mark.parametrize("instance", ("single_instance", "multi_instance"))
    def test_upgrade(self, instance):
        """It should run the mysql upgrade command."""
        self.mocked_run_sync.return_value = iter(())
        getattr(self, instance).upgrade()
        suffix = ".instance1" if instance == "multi_instance" else ""
        self.mocked_run_sync.assert_called_once_with(
            f"$(readlink -f /usr/local/bin/mysql_upgrade) --socket /run/mysqld/mysqld{suffix}.sock --force",
            print_output=True,
        )

    def test_get_replication_info_ok(self):
        """It should return a ReplicationInfo instance with the proper data."""
        mocked_status = json.loads(get_fixture_path("mysql", "single_show_slave_status.json").read_text())
        self.mocked_cursor.execute.return_value = 1
        self.mocked_cursor.fetchone.return_value = mocked_status

        info = self.single_instance.get_replication_info()

        assert info.primary == "host1.example.org"
        assert info.binlog == "host1-bin.001234"
        assert info.position == 123456782
        self.mocked_cursor.execute.assert_called_once_with("SHOW SLAVE STATUS")
        self.mocked_cursor.fetchone.assert_called_once_with()

    def test_get_replication_info_raise(self):
        """It should raise a MysqlError if unable to get the replication information."""
        mocked_status = json.loads(get_fixture_path("mysql", "single_show_slave_status.json").read_text())
        del mocked_status["Master_Host"]  # Make it invalid
        self.mocked_cursor.execute.return_value = 1
        self.mocked_cursor.fetchone.return_value = mocked_status

        with pytest.raises(mysql.MysqlError, match="Could not find the replication position"):
            self.single_instance.get_replication_info()

        self.mocked_cursor.execute.assert_called_once_with("SHOW SLAVE STATUS")
        self.mocked_cursor.fetchone.assert_called_once_with()

    def test_primary_ok(self):
        """It should return the hostname of the primary host for this host."""
        mocked_status = json.loads(get_fixture_path("mysql", "single_show_slave_status.json").read_text())
        self.mocked_cursor.execute.return_value = 1
        self.mocked_cursor.fetchone.return_value = mocked_status

        assert self.single_instance.primary == "host1.example.org"

        self.mocked_cursor.execute.assert_called_once_with("SHOW SLAVE STATUS")
        # Ensure the caching of the result works
        self.single_instance.primary  # pylint: disable=pointless-statement
        assert self.mocked_cursor.execute.call_count == 1
        assert self.mocked_cursor.fetchone.call_count == 1

    @pytest.mark.parametrize("no_content", (False, True))
    def test_primary_raise(self, no_content):
        """It should raise a MysqlError if there is no primary or is run on a master."""
        mocked_status = json.loads(get_fixture_path("mysql", "single_show_slave_status.json").read_text())
        del mocked_status["Master_Host"]  # Make it invalid
        self.mocked_cursor.execute.return_value = 0 if no_content else 1
        self.mocked_cursor.fetchone.return_value = None if no_content else mocked_status

        with pytest.raises(mysql.MysqlError, match="Unable to retrieve master host"):
            self.single_instance.primary  # pylint: disable=pointless-statement

        if not no_content:
            self.mocked_cursor.fetchone.assert_called_once_with()

    def test_prep_src_for_cloning(self):
        """It should run the preparation commands before cloning."""
        mocked_status = json.loads(get_fixture_path("mysql", "single_show_slave_status.json").read_text())
        self.mocked_cursor.execute.side_effect = [0, 0, 1, 0]
        self.mocked_cursor.fetchone.return_value = mocked_status
        self.mocked_run_sync.return_value = [(nodeset("single1"), iter(()))]

        info = self.single_instance.prep_src_for_cloning()

        assert info.primary == "host1.example.org"
        self.mocked_run_sync.assert_called_once_with("/usr/bin/systemctl stop mariadb.service")
        self.mocked_cursor.execute.assert_any_call("STOP SLAVE", None)
        self.mocked_cursor.execute.assert_any_call("SHOW SLAVE STATUS")
        self.mocked_cursor.fetchone.assert_called_once_with()

    def test_set_replication_parameters(self):
        """It should set the replication to the given parameters."""
        info = mysql.ReplicationInfo(
            primary="host1.example.org", binlog="host1-bin.001234", position=123456782, port=3306
        )
        self.mocked_cursor.execute.return_value = 0

        self.multi_instance.set_replication_parameters(replication_info=info, user="user", password="dummy")

        call_args = self.mocked_cursor.execute.call_args_list[0].args
        for part in (
            "CHANGE MASTER TO",
            "master_host=%(primary)s",
            "master_port=%(port)s",
            "master_ssl=1",
            "master_log_file=%(binlog)s",
            "master_log_pos=%(position)s",
            "master_user=%(user)s",
            "password=%(password)s",
        ):
            assert part in call_args[0]

        expected = {
            "primary": "host1.example.org",
            "port": 3306,
            "binlog": "host1-bin.001234",
            "position": 123456782,
            "user": "user",
            "password": "dummy",
        }
        for key, value in expected.items():
            assert call_args[1][key] == value

    def test_post_clone_reset_with_slave_stopped(self):
        """It should start mysql with slave stopped and reset all slave information."""
        self.mocked_run_sync.side_effect = [[(nodeset("single1"), iter(()))]] * 3
        self.mocked_cursor.execute.return_value = 0

        self.single_instance.post_clone_reset_with_slave_stopped()

        calls = self.mocked_run_sync.call_args_list
        assert calls[0][0][0].endswith("chown -R mysql:mysql /srv/sqldata")
        assert calls[0][0][1].endswith('set-environment MYSQLD_OPTS="--skip-slave-start"')
        assert calls[1][0][0].endswith("systemctl start mariadb.service")
        self.mocked_cursor.execute.assert_any_call("STOP SLAVE", None)
        self.mocked_cursor.execute.assert_any_call("RESET SLAVE ALL", None)

    @mock.patch("spicerack.mysql.sleep", return_value=None)
    def test_resume_replication(self, mocked_sleep):
        """It should start mysql, upgrade it, restart it and resume the replication."""
        self.mocked_run_sync.side_effect = [[(nodeset("single1"), iter(()))]] * 4
        self.mocked_cursor.execute.return_value = 0

        self.single_instance.resume_replication()

        calls = self.mocked_run_sync.call_args_list
        assert calls[0][0][0].endswith('set-environment MYSQLD_OPTS="--skip-slave-start"')
        assert calls[1][0][0].endswith("systemctl start mariadb.service")
        assert calls[2][0][0].endswith(
            "$(readlink -f /usr/local/bin/mysql_upgrade) --socket /run/mysqld/mysqld.sock --force"
        )
        assert calls[3][0][0].endswith("systemctl restart mariadb.service")
        self.mocked_cursor.execute.assert_any_call("START SLAVE", None)
        mocked_sleep.assert_called_once_with(1)

    @pytest.mark.parametrize("threshold", (0, Decimal("0.1234"), 0.5))
    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_wait_for_replication_already_ok(self, mocked_sleep, threshold):
        """If the replication is already in sync it should return immediately."""
        self.mocked_cursor.execute.side_effect = [1, 0]
        self.mocked_cursor.fetchone.return_value = {"lag": Decimal("0.1234")}
        if threshold:
            self.single_instance.wait_for_replication(threshold)
        else:
            threshold = 1.0
            self.multi_instance.wait_for_replication()

        assert self.mocked_cursor.execute.call_count == 2
        assert self.mocked_pymysql.connect.call_args.kwargs["database"] == "heartbeat"
        self.mocked_cursor.fetchone.assert_called_once_with()
        mocked_sleep.assert_not_called()

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_wait_for_replication_sleep_ok(self, mocked_sleep):
        """If the replication is not yet in sync it should wait until it gets in sync, within the timeout."""
        self.mocked_cursor.execute.side_effect = [1, 0] * 6
        lags = [{"lag": Decimal("1.2345")}] * 5
        lags.append({"lag": Decimal("0.1234")})
        self.mocked_cursor.fetchone.side_effect = lags

        self.single_instance.wait_for_replication()

        assert mocked_sleep.call_count == 5

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_wait_for_replication_fail(self, mocked_sleep):
        """If the replication is not in sync within the timeout it should raise a MysqlReplagError exception."""
        self.mocked_cursor.execute.side_effect = [1, 0] * 480
        self.mocked_cursor.fetchone.side_effect = [{"lag": Decimal("1.2345")}] * 480

        with pytest.raises(
            mysql.MysqlReplagError,
            match=re.escape("Replication lag higher than the threshold (1.0s): 1.2345s"),
        ):
            self.single_instance.wait_for_replication()

        assert mocked_sleep.call_count == 479

    def test_replication_lag_ok(self):
        """It should return the current replication lag as float."""
        self.mocked_cursor.execute.side_effect = [1, 0]
        self.mocked_cursor.fetchone.return_value = {"lag": Decimal("0.1234")}
        assert self.single_instance.replication_lag() == Decimal("0.1234")
        assert "TIMESTAMPDIFF" in self.mocked_cursor.execute.call_args_list[0].args[0]

    def test_replication_lag_no_data(self):
        """It should raise a MysqlError if the query returns no rows."""
        self.mocked_cursor.execute.return_value = 0
        with pytest.raises(mysql.MysqlError, match="The replication lag query returned no data"):
            self.single_instance.replication_lag()

        self.mocked_cursor.fetchone.assert_not_called()

    @pytest.mark.parametrize(
        "row",
        (
            {"lag": None},
            {"invalid": Decimal(0.1234)},
        ),
    )
    def test_replication_lag_no_lag(self, row):
        """It should raise a MysqlLegacyError if the returned lag is None or invalid."""
        self.mocked_cursor.execute.side_effect = [1, 0]
        self.mocked_cursor.fetchone.return_value = row
        error_message = f"Unable to get lag information from: {row}"

        with pytest.raises(mysql.MysqlError, match=re.escape(error_message)):
            self.single_instance.replication_lag()

        self.mocked_cursor.fetchone.assert_called_once_with()


class TestMysqlRemoteHosts:
    """Test class for the MysqlRemoteHosts class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_run_sync = mock.Mock()
        self.config = Config(get_fixture_path("remote", "config.yaml"))

        remote_hosts = RemoteHosts(self.config, nodeset("host[1-9]"), dry_run=False)
        remote_hosts.run_sync = self.mocked_run_sync
        self.mysql_remote_hosts = mysql.MysqlRemoteHosts(remote_hosts)

        remote_host = RemoteHosts(self.config, nodeset("host1"), dry_run=False)
        remote_host.run_sync = self.mocked_run_sync
        self.mysql_remote_host = mysql.MysqlRemoteHosts(remote_host)

    def test_iter(self):
        """It should iterate the instance yielding instances of the same class with one host."""
        expected = ["host1", "host2", "host3", "host4", "host5", "host6", "host7", "host8", "host9"]
        for index, mysql_remote_host in enumerate(self.mysql_remote_hosts):
            assert isinstance(mysql_remote_host, mysql.MysqlRemoteHosts)
            assert str(mysql_remote_host) == expected[index]

    def test_run_query(self):
        """Calling run_query() should run the given query in the target hosts."""
        self.mocked_run_sync.return_value = [(nodeset("host[1-9]"), MsgTreeElem(b"result", parent=MsgTreeElem()))]
        results = self.mysql_remote_hosts.run_query("SELECT 'result'")
        assert [(str(host), msg.message().decode()) for host, msg in results] == [("host[1-9]", "result")]
        self.mocked_run_sync.assert_called_once_with(
            "/usr/local/bin/mysql --skip-ssl --skip-column-names --batch -e \"SELECT 'result'\"",
            print_progress_bars=False,
            print_output=False,
        )

    def test_list_host_instances_no_instance(self):
        """It should return an empty list if there are no instances on the host."""
        self.mocked_run_sync.side_effect = [
            [],  # Empty ls result
            RemoteExecutionError(retcode=1, message="error", results=iter(())),  # Empty grep result
        ]
        instances = self.mysql_remote_host.list_hosts_instances()
        assert isinstance(instances, list)
        assert not instances

    def test_list_host_instances_single(self):
        """It should return a list with just one Instance object for the single instance."""
        self.mocked_run_sync.side_effect = [[], []]  # Empty ls and successful grep -q
        instances = self.mysql_remote_host.list_hosts_instances()
        assert len(instances) == 1
        assert isinstance(instances[0], mysql.Instance)
        assert str(instances[0].host) == "host1"
        assert instances[0].name == ""

    def test_list_host_instances_multi(self):
        """It should return a list with all the instances of a multi-instance host."""
        configs = b"instance1.cnf\ninstance2.cnf"
        self.mocked_run_sync.return_value = [(nodeset("host1"), MsgTreeElem(configs, parent=MsgTreeElem()))]
        instances = self.mysql_remote_host.list_hosts_instances()
        assert len(instances) == 2
        for instance in instances:
            assert isinstance(instance, mysql.Instance)
            assert str(instance.host) == "host1"

        assert instances[0].name == "instance1"
        assert instances[1].name == "instance2"

    def test_list_host_instances_not_single_host(self):
        """It should raise a NotImplementedError if the MysqlRemoteHosts instance has multiple hosts."""
        with pytest.raises(NotImplementedError, match="Only single host are supported at this time"):
            self.mysql_remote_hosts.list_hosts_instances()

    def test_list_host_instances_grouped(self):
        """It should raise a NotImplementedError if the grouped parameter is set to True."""
        with pytest.raises(NotImplementedError, match="Grouping and parallelization are not supported at this time"):
            self.mysql_remote_host.list_hosts_instances(group=True)


class TestMysql:
    """Mysql class tests."""

    @mock.patch("spicerack.remote.transports", autospec=True)
    def setup_method(self, _, mocked_transports):
        """Initialize the test environment for Mysql."""
        # pylint: disable=attribute-defined-outside-init
        self.config = Config(get_fixture_path("remote", "config.yaml"))
        self.mocked_transports = mocked_transports
        self.mocked_remote = mock.MagicMock(spec_set=Remote)
        self.mysql = mysql.Mysql(self.mocked_remote, dry_run=False)

    def test_get_dbs(self):
        """It should return and instance of MysqlRemoteHosts for the matching hosts."""
        self.mysql.get_dbs("query")
        self.mocked_remote.query.assert_called_once_with("query")

    @pytest.mark.parametrize(
        "kwargs, query, match",
        (
            ({}, "A:db-core", "db10[01-99],db20[01-99]"),
            ({"datacenter": "eqiad"}, "A:db-core and A:eqiad", "db10[01-99]"),
            (
                {"section": "s1"},
                "A:db-core and A:db-section-s1",
                "db10[01-10],db20[01-10]",
            ),
            (
                {"replication_role": "master"},
                "A:db-core and A:db-role-master",
                ",".join([EQIAD_CORE_MASTERS_QUERY, CODFW_CORE_MASTERS_QUERY]),
            ),
            (
                {"datacenter": "eqiad", "section": "s1"},
                "A:db-core and A:eqiad and A:db-section-s1",
                "db10[01-10]",
            ),
            (
                {"datacenter": "eqiad", "replication_role": "master"},
                "A:db-core and A:eqiad and A:db-role-master",
                EQIAD_CORE_MASTERS_QUERY,
            ),
            (
                {"datacenter": "eqiad", "replication_role": "master", "excludes": ("s2",)},
                "A:db-core and A:eqiad and not A:db-section-s2 and A:db-role-master",
                "db10[01-10]",
            ),
            (
                {"section": "s1", "replication_role": "master"},
                "A:db-core and A:db-section-s1 and A:db-role-master",
                "db1001,db2001",
            ),
            (
                {"datacenter": "eqiad", "section": "s1", "replication_role": "master"},
                "A:db-core and A:eqiad and A:db-section-s1 and A:db-role-master",
                "db1001",
            ),
        ),
    )
    def test_get_core_dbs_ok(self, kwargs, query, match):
        """It should return the right DBs based on the parameters."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset(match))
        self.mysql.get_core_dbs(**kwargs)
        self.mocked_remote.query.assert_called_once_with(query)

    @pytest.mark.parametrize(
        "kwargs",
        (
            {"datacenter": "invalid"},
            {"section": "invalid"},
            {"replication_role": "invalid"},
            {"excludes": ("invalid",)},
        ),
    )
    def test_get_core_dbs_fail(self, kwargs):
        """It should raise MysqlError if called with invalid parameters."""
        key = list(kwargs.keys())[0]
        message = f"Got invalid {key}"
        with pytest.raises(mysql.MysqlError, match=message):
            self.mysql.get_core_dbs(**kwargs)

        assert not self.mocked_remote.query.called

    def test_get_core_dbs_fail_sanity_check(self):
        """It should raise MysqlError if matching an invalid number of hosts when looking for masters."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db1001"))
        with pytest.raises(mysql.MysqlError, match="Matched 1 masters, expected 11"):
            self.mysql.get_core_dbs(datacenter="eqiad", replication_role="master")

        assert self.mocked_remote.query.called

    @pytest.mark.parametrize("mode, value", (("readonly", b"1"), ("readwrite", b"0")))
    def test_set_core_masters_readonly(self, mode, value, caplog):
        """It should set the masters as read-only/read-write."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db10[01-11]"))
        mock_cumin(self.mocked_transports, 0, retvals=[[("db10[01-11]", value)]])
        with caplog.at_level(logging.DEBUG):
            getattr(self.mysql, "set_core_masters_" + mode)("eqiad")
        assert "SET GLOBAL read_only=" + value.decode() in caplog.text

    @pytest.mark.parametrize("readonly, reply", ((True, b"1"), (False, b"0")))
    def test_verify_core_masters_readonly_ok(self, readonly, reply, caplog):
        """Should verify that the masters have the intended read-only value."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db10[01-11]"))
        mock_cumin(self.mocked_transports, 0, retvals=[[("db10[01-11]", reply)]])
        with caplog.at_level(logging.DEBUG):
            self.mysql.verify_core_masters_readonly("eqiad", readonly)
        assert "SELECT @@global.read_only" in caplog.text

    def test_verify_core_masters_readonly_fail(self):
        """Should raise MysqlError if some masters do not have the intended read-only value."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db10[01-11]"))
        mock_cumin(
            self.mocked_transports,
            0,
            retvals=[[("db1001", b"0"), ("db10[02-11]", b"1")]],
        )
        with pytest.raises(
            mysql.MysqlError,
            match="Verification failed that core DB masters",
        ):
            self.mysql.verify_core_masters_readonly("eqiad", True)

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_check_core_masters_in_sync_ok(self, mocked_sleep):
        """Should check that all core masters are in sync with the master in the other DC."""
        hosts = nodeset(EQIAD_CORE_MASTERS_QUERY)
        self.mocked_remote.query.side_effect = [RemoteHosts(self.config, nodeset(host)) for host in hosts] * 2
        retvals = [[(host, b"2018-09-06T10:00:00.000000")] for host in hosts]  # first heartbeat
        retvals += [[(host, b"2018-09-06T10:00:01.000000")] for host in hosts]  # second heartbeat
        mock_cumin(self.mocked_transports, 0, retvals=retvals)
        self.mysql.check_core_masters_in_sync("eqiad", "codfw")
        assert not mocked_sleep.called

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_check_core_masters_in_sync_fail_heartbeat(self, mocked_sleep):
        """Should raise MysqlError if unable to get the heartbeat from the current master."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db1001"))
        mock_cumin(self.mocked_transports, 0, retvals=[])
        with pytest.raises(mysql.MysqlError, match="Unable to get heartbeat from master"):
            self.mysql.check_core_masters_in_sync("eqiad", "codfw")
        assert not mocked_sleep.called

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_check_core_masters_in_sync_not_in_sync(self, mocked_sleep):
        """Should raise MysqlError if a master is not in sync with the one in the other DC."""
        hosts = nodeset(EQIAD_CORE_MASTERS_QUERY)
        self.mocked_remote.query.side_effect = [RemoteHosts(self.config, nodeset(host)) for host in hosts] + [
            RemoteHosts(self.config, nodeset("db1001"))
        ] * 3
        retvals = [[(host, b"2018-09-06T10:00:00.000000")] for host in hosts]  # first heartbeat
        retvals += [[("db1001", b"2018-09-06T10:00:00.000000")]] * 3  # 3 failed retries of second heartbeat
        mock_cumin(self.mocked_transports, 0, retvals=retvals)
        with pytest.raises(
            mysql.MysqlError,
            match=r"Heartbeat from master db1001 for section .* not yet in sync",
        ):
            self.mysql.check_core_masters_in_sync("eqiad", "codfw")

        assert mocked_sleep.called

    def test_get_core_masters_heartbeats_wrong_data(self):
        """Should raise MysqlError if unable to convert the heartbeat into a datetime."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db1001"))
        mock_cumin(
            self.mocked_transports,
            0,
            retvals=[[("db1001", b"2018-09-06-10:00:00.000000")]],
        )
        with pytest.raises(mysql.MysqlError, match="Unable to convert heartbeat"):
            self.mysql.get_core_masters_heartbeats("eqiad", "codfw")

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_check_core_masters_heartbeats_fail(self, mocked_sleep):
        """Should raise MysqlError if unable to get the heartbeat from the master."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db1001"))
        mock_cumin(self.mocked_transports, 0, retvals=[])
        with pytest.raises(mysql.MysqlError, match="Unable to get heartbeat from master"):
            self.mysql.check_core_masters_heartbeats("eqiad", "codfw", {"s1": datetime.utcnow()})

        assert mocked_sleep.called


class TestMysqlClient:
    """MysqlClient class tests."""

    @mock.patch("spicerack.mysql.Connection", autospec=True)
    def test_connect_default(self, mocked_pymsql_connection):
        """It should call pymysql with the correct parameters."""
        my = mysql.MysqlClient(dry_run=False)
        with my.connect() as conn:
            assert conn == mocked_pymsql_connection.return_value
            call_args = mocked_pymsql_connection.call_args.kwargs
            # Ensure close is not called before context manager exits
            conn.close.assert_not_called()  # pylint: disable=maybe-no-member

        conn.close.assert_called_once_with()  # pylint: disable=maybe-no-member

        # Ensure the default args were passed
        if call_args.get("host") == "clouddb1001":
            assert call_args["read_default_group"] == "clientlabsdb"
        else:
            assert call_args["read_default_group"] == "client"

        assert call_args["charset"] == "utf8mb4"
        assert call_args["cursorclass"] == DictCursor
        assert call_args["read_default_file"].endswith("/.my.cnf")
        assert call_args["ssl"] == {"ca": WMF_CA_BUNDLE_PATH}

    @pytest.mark.parametrize(
        "kwargs",
        (
            {"host": "db9999"},
            {"charset": "ascii"},
            {"charset": ""},
            {"read_default_file": "/my.cnf"},
            {"read_default_file": None, "read_default_group": None},
            {"read_default_group": "client_test"},
            {"host": "clouddb1001"},
            {"ssl": {}},
            {"ssl": {1: 3}},
        ),
    )
    @mock.patch("spicerack.mysql.Connection", autospec=True)
    def test_connect(self, mocked_pymsql_connection, kwargs):
        """It should call pymysql with the correct parameters."""
        my = mysql.MysqlClient(dry_run=False)
        with my.connect(**kwargs) as conn:
            assert conn == mocked_pymsql_connection.return_value
            call_args = mocked_pymsql_connection.call_args.kwargs
            # Ensure close is not called before context manager exits
            conn.close.assert_not_called()  # pylint: disable=maybe-no-member

        conn.close.assert_called_once_with()  # pylint: disable=maybe-no-member
        # Ensure the args we passed were passed along
        for key, value in kwargs.items():
            assert call_args[key] == value

    @pytest.mark.parametrize(
        "dry_run, read_only, transaction_ro",
        (
            (False, False, False),
            (False, True, True),
            (True, False, True),
            (True, True, True),
        ),
    )
    @mock.patch("spicerack.mysql.Connection", autospec=True)
    def test_connect_read_only(self, mocked_pymsql_connection, dry_run, read_only, transaction_ro):
        """It should start a read-only transaction if either dry-run or read-only are set."""
        my = mysql.MysqlClient(dry_run=dry_run)
        with my.connect(read_only=read_only) as conn:
            assert conn == mocked_pymsql_connection.return_value
            execute = conn.cursor.return_value.__enter__.return_value.execute  # pylint: disable=no-member
            print(mocked_pymsql_connection.mock_calls)
            if transaction_ro:
                execute.assert_called_once_with("SET SESSION TRANSACTION READ ONLY")
            else:
                execute.assert_not_called()



"""Call another cookbook cookbook."""

import argparse

from spicerack.cookbook import CookbookBase, CookbookRunnerBase


class CallAnotherCookbook(CookbookBase):
    """A cookbook that calls another cookbook."""

    owner_team = "team1"

    def argument_parser(self):
        """As defined by the parent class."""
        parser = argparse.ArgumentParser("Argparse")
        parser.add_argument("cookbook", help="The path of the cookbook to execute.")
        parser.add_argument("--raises", action="store_true", help="Set raises=True in run_cookbook")
        parser.add_argument("--confirm", action="store_true", help="Set confirm=True in run_cookbook")
        return parser

    def get_runner(self, args):
        """As required by the parent class."""
        return ExampleRunner(args, self.spicerack)


class ExampleRunner(CookbookRunnerBase):
    """As required by spicerack._cookbook."""

    def __init__(self, args, spicerack):
        """Initialize the instance."""
        self.args = args
        self.spicerack = spicerack

    def run(self):
        """As required by the parent class."""
        return self.spicerack.run_cookbook(self.args.cookbook, [], raises=self.args.raises, confirm=self.args.confirm)



"""Cookbook module tests."""

from unittest import mock

import pytest

from spicerack import Spicerack, cookbook


class Cookbook(cookbook.CookbookBase):
    """Cookbook class used in the tests."""

    def get_runner(self, _):
        """As required by Spicerack APIs."""
        return CookbookRunner()


class CookbookRunner(cookbook.CookbookRunnerBase):
    """Cookbook runner class used in the tests."""

    def run(self):
        """As required by Spicerack APIs."""


class TestCookbookBase:
    """Test class for the CookbookBase class."""

    def setup_method(self):
        """Initialize the test instance."""
        # pylint: disable=attribute-defined-outside-init
        self.cookbook = Cookbook(mock.MagicMock(spec_set=Spicerack))

    def test_argument_parser_empty_ok(self):
        """If using the parent class argument parser without options it should return an empty argument parser."""
        parser = self.cookbook.argument_parser()
        parsed_args = parser.parse_args([])
        assert not hasattr(parsed_args, "reason")
        assert not hasattr(parsed_args, "task_id")

    @pytest.mark.parametrize("argument_task_required", (False, True))
    @pytest.mark.parametrize("arg_name", ("-t", "--task-id"))
    def test_argument_parser_task(self, arg_name, argument_task_required):
        """If the task ID argument is requested, either optionally or mandatory, it should be added."""
        self.cookbook.argument_task_required = argument_task_required
        parser = self.cookbook.argument_parser()
        args = [arg_name, "T12345"]
        parsed_args = parser.parse_args(args)
        assert parsed_args.task_id == "T12345"
        assert not hasattr(parsed_args, "reason")

    @pytest.mark.parametrize("argument_reason_required", (False, True))
    @pytest.mark.parametrize("arg_name", ("-r", "--reason"))
    def test_argument_parser_reason(self, arg_name, argument_reason_required):
        """If the reason argument is requested, either optionally or mandatory, it should be added."""
        self.cookbook.argument_reason_required = argument_reason_required
        parser = self.cookbook.argument_parser()
        args = [arg_name, "Some reason"]
        parsed_args = parser.parse_args(args)
        assert parsed_args.reason == "Some reason"
        assert not hasattr(parsed_args, "task_id")

    @pytest.mark.parametrize("argument_task_required", (False, True))
    @pytest.mark.parametrize("task_name", ("-t", "--task-id"))
    @pytest.mark.parametrize("argument_reason_required", (False, True))
    @pytest.mark.parametrize("reason_name", ("-r", "--reason"))
    def test_argument_parser_reason_task(
        self, reason_name, argument_reason_required, task_name, argument_task_required
    ):
        """If both the task ID and the reason argument are requested, they should be added."""
        self.cookbook.argument_task_required = argument_task_required
        self.cookbook.argument_reason_required = argument_reason_required
        parser = self.cookbook.argument_parser()
        args = [reason_name, "Some reason", task_name, "T12345"]
        parsed_args = parser.parse_args(args)
        assert parsed_args.reason == "Some reason"
        assert parsed_args.task_id == "T12345"

    def test_argument_parser_reason_task_optional(self):
        """If the task ID and the reason argument are requested as optional, it shouldn't error if they are not set."""
        self.cookbook.argument_task_required = False
        self.cookbook.argument_reason_required = False
        parser = self.cookbook.argument_parser()
        parsed_args = parser.parse_args([])
        assert parsed_args.reason is None
        assert parsed_args.task_id is None

    @pytest.mark.parametrize(
        "kwargs",
        (
            {"argument_task_required": True},
            {"argument_reason_required": True},
            {"argument_task_required": True, "argument_reason_required": True},
        ),
    )
    def test_argument_parser_missing_mandatory_argument(self, kwargs):
        """If the task ID or the reason argument are requested as mandatory, if the args are not set should raise."""
        for key, value in kwargs.items():
            setattr(self.cookbook, key, value)
        parser = self.cookbook.argument_parser()
        with pytest.raises(SystemExit, match="2"):
            parser.parse_args([])

    @pytest.mark.parametrize(
        "task",
        (
            "",
            "T",
            "123",
            "T123456789",
            "A",
            "A123",
        ),
    )
    @pytest.mark.parametrize("arg_name", ("-t", "--task-id"))
    def test_argument_parser_task_invalid(self, arg_name, task):
        """If the provided task ID doesn't match a Phabricator task ID format it should raise."""
        self.cookbook.argument_task_required = False
        parser = self.cookbook.argument_parser()
        with pytest.raises(SystemExit, match="2"):
            parser.parse_args([arg_name, task])

    @pytest.mark.parametrize(
        "reason",
        (
            "",
            'reason "with quotes"',
            '"',
        ),
    )
    @pytest.mark.parametrize("arg_name", ("-r", "--reason"))
    def test_argument_parser_reason_invalid(self, arg_name, reason):
        """If the provided reason is not valid it should raise."""
        self.cookbook.argument_reason_required = False
        parser = self.cookbook.argument_parser()
        with pytest.raises(SystemExit, match="2"):
            parser.parse_args([arg_name, reason])



"""MySQL shell module."""

import logging
from collections.abc import Generator, Iterator
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime, timedelta
from decimal import Decimal
from enum import Enum
from pathlib import Path
from time import sleep
from typing import Any, Optional, Union

from ClusterShell.MsgTree import MsgTreeElem
from cumin import NodeSet
from cumin.transports import Command
from pymysql.connections import Connection
from pymysql.cursors import DictCursor
from wmflib.constants import CORE_DATACENTERS
from wmflib.interactive import ask_confirmation

from spicerack.constants import WMF_CA_BUNDLE_PATH
from spicerack.decorators import retry
from spicerack.exceptions import SpicerackError
from spicerack.remote import Remote, RemoteExecutionError, RemoteHosts, RemoteHostsAdapter

REPLICATION_ROLES: tuple[str, ...] = ("master", "slave", "standalone")
"""Valid replication roles."""
CORE_SECTIONS: tuple[str, ...] = (
    "s6",
    "s5",
    "s2",
    "s7",
    "s3",
    "s8",
    "s4",
    "s1",
    "x1",
    "es6",
    "es7",
)
"""Valid MySQL RW core sections (external storage RO, parser cache, x2 and misc sections are not included here).
They are ordered from less impactful if anything goes wrong to most impactful.
"""

logger = logging.getLogger(__name__)


class MysqlError(SpicerackError):
    """Custom exception class for errors of this module."""


class MysqlReplagError(MysqlError):
    """Custom exception class for errors related to replag in this module."""


# TODO: use StrEnum once on 3.11+
class MasterUseGTID(Enum):
    """Describe the possible values for the MASTER_USE_GTID option.

    See Also:
        https://mariadb.com/kb/en/change-master-to/#master_use_gtid

    """

    CURRENT_POS = "current_pos"
    """Replicate in GTID mode and use gtid_current_pos as the position to start downloading transactions."""
    SLAVE_POS = "slave_pos"
    """Replicate in GTID mode and use gtid_slave_pos as the position to start downloading transactions."""
    NO = "no"
    """Don't replicate in GTID mode."""


@dataclass
class ReplicationInfo:
    """Represent the minimum replication information needed to restore a replication from a given point.

    Arguments:
        primary: the FQDN of the primary host from where to replicate from.
        binlog: the binlog file to replicate from.
        position: the binlog position to replicate from.
        port: the port of the master from where to replicate from.

    """

    primary: str
    binlog: str
    position: int
    port: int


class MysqlClient:
    """Connect to Mysql instances with a native Mysql client (pymysql).

    Caution:
        This class only has DRY-RUN support for DML sql operations.

    """

    def __init__(self, *, dry_run: bool = True) -> None:
        """Initialize the instance.

        Arguments:
            dry_run: whether this is a DRY-RUN.

        """
        self._dry_run = dry_run

    @contextmanager
    def connect(self, *, read_only: bool = False, **kwargs: Any) -> Generator:
        """Context-manager for a mysql connection to a remote host.

        Caution:
            Read-only support is limited to DML sql operations.

        Important:
            * By default autocommit is off and the commit of changes is the caller's responsibility.
            * The caller should also take care of rolling back transactions on error as appropriate.

        Arguments:
            read_only: True if this connection should use read-only transactions. **Note**: This parameter has no
                effect if DRY-RUN is set, it will be forced to True.
            **kwargs: Options passed directly to :py:class:`pymysql.connections.Connection`. If not set some settings
                will be set to default values:

                    * read_default_file: ``~/.my.cnf``. Set to :py:data:`None` to disable it.
                    * read_default_group: If not specified, it will be set to ``client`` or ``labsdbclient`` based on
                      the hostname.
                    * ssl: uses the puppet CA. Set to ``{}`` to disable.
                    * cursorclass: :py:class:`pymysql.cursors.DictCursor`

        Yields:
            :py:class:`pymysql.connections.Connection`: a context-managed mysql connection.

        Raises:
            pymysql.err.MySQLError: if unable to create the connection or set the connection to read only.

        """
        default_group = "client"
        if kwargs.get("host", "").startswith("clouddb"):
            default_group = "clientlabsdb"

        params: dict[str, Any] = {
            "charset": "utf8mb4",
            "cursorclass": DictCursor,
            "read_default_file": str(Path("~/.my.cnf").expanduser()),
            "read_default_group": default_group,
            "ssl": {"ca": WMF_CA_BUNDLE_PATH},
        }
        params.update(kwargs)

        conn = Connection(**params)

        if read_only or self._dry_run:
            # FIXME: read-only support is limited to DML sql statements.
            # https://phabricator.wikimedia.org/T254756 is needed to do this better.
            with conn.cursor() as cursor:
                _ = cursor.execute("SET SESSION TRANSACTION READ ONLY")

        try:
            yield conn
        # Not catching exceptions and rolling back, as that restricts the client code
        # in how it does error handling.
        finally:
            conn.close()


class Instance:
    """Class to manage MariaDB single intances and multiinstances."""

    def __init__(self, host: RemoteHosts, *, name: str = "") -> None:
        """Initialize the instance.

        Arguments:
            host: the RemoteHosts instance that contains this MariaDB SingleInstance.
            name: the name of the instance in a multiinstance context. Leave it empty for single instances.

        """
        if len(host) > 1:
            raise NotImplementedError("Only single hosts are currently supported. Got {len(host)}.")

        self.host = host
        self.name = name
        self._mysql = MysqlClient(dry_run=host.dry_run)
        self._primary = ""
        self._mysql_bin = "/usr/local/bin/mysql"

        if self.name:
            self._sock = f"/run/mysqld/mysqld.{self.name}.sock"
            self._service = f"mariadb@{self.name}.service"
            self._data_dir = f"/srv/sqldata.{self.name}"
        else:
            self._sock = "/run/mysqld/mysqld.sock"
            self._service = "mariadb.service"
            self._data_dir = "/srv/sqldata"

    def __str__(self) -> str:
        """Return a string representation of the instance.

        Returns:
            the FQDN and name of the instance, when present.

        """
        name = self.name if self.name else "single-instance"
        return f"{self.host} ({name})"

    @contextmanager
    def cursor(self, **kwargs: Any) -> Generator:
        """Context manager to get an open connection and cursor against the current instance.

        Caution:
            DRY-RUN and read-only support is limited to DML sql operations.

        Examples:
            * Iterate directly the cursor
              ::

                >>> with instance.cursor(database="mydb") as (connection, cursor):
                >>>     count = cursor.execute("SELECT * FROM mytable WHERE a = %s and b = %s", ("value", 10))
                >>>     for row in cursor:
                >>>         # example row: {'a': 'value', 'b': 10, 'c': 1}
                >>>
                >>>     instance.check_warnings(cursor)

            * Get all the results
              ::

                >>> with instance.cursor(database="mydb") as (connection, cursor):
                >>>     count = cursor.execute("SELECT * FROM mytable WHERE a = %s and b = %s", ("value", 10))
                >>>     results = cursor.fetchall()
                >>>     instance.check_warnings(cursor)
                >>>
                >>> count
                2
                >>> results
                [{'a': 'value', 'b': 10, 'c': 1}, {'a': 'value', 'b': 10, 'c': 2}]

        Arguments:
            **kwargs: arbitrary arguments that are passed to the :py:class:`spicerack.mysql.MysqlClient.connect`
                method. See its documentation for the available arguments and their default values.

        Yields:
            A two-element tuple with the :py:class:`pymysql.connections.Connection` as first element and the cursor
            object as second element. The cursor object is one of :py:mod:`pymysql.cursors`.

        Raises:
            pymysql.err.MySQLError: or any more specific exception that inherits from this one on error.

        """
        # TODO: make the Instance class be aware of the port to use to connect to mysql, ideally allowing to use the
        # admin port if needed.
        kwargs["host"] = str(self.host)
        with self._mysql.connect(**kwargs) as connection:
            with connection.cursor() as cursor:
                yield connection, cursor

    def check_warnings(self, cursor: DictCursor) -> None:
        """It will check if there is any warning in the cursor for the last query and ask the user what to do.

        If any warning is found they will be logged to console and logfile and the user will be prompted what to do.

        Arguments:
            cursor: the cursor object with which the last query was made.

        """
        num_warnings = cursor.execute("SHOW WARNINGS")
        if not num_warnings:
            return

        warnings = cursor.fetchall()
        for warning in warnings:
            logger.warning("[%s] %s: %s", warning["Level"], warning["Code"], warning["Message"])

        ask_confirmation("The above warnings were raised during the last query, do you want to proceed anyway?")

    def execute(
        self,
        query: str,
        query_parameters: Union[None, tuple, list, dict] = None,
        *,
        is_safe: bool = False,
        **kwargs: Any,
    ) -> int:
        """Execute a query that returns no data without giving access to the connection or cursor objects.

        Caution:
            DRY-RUN and read-only support is limited to DML sql operations when ``is_safe`` is set to :py:data:`True`.

        Note:
            If any warning is issued by the database they will be logged and the user prompted what to do.

        Examples:
            ::

                >>> query = "INSERT INTO mytable VALUES (%(a)s, %(b)s)"
                >>> params = {"a": "value", "b": 10}
                >>> num_rows = instance.execute(query, params, database="mydb")

        Arguments:
            query: the query to execute, with eventual placeholders (``%s`` or ``%(name)s``).
            query_parameters: the query parameters to inject into the query, a :py:class:`tuple` or :py:class:`list` in
                case ``%s`` placeholders were used or a :py:class:`dict` in case ``%(name)s`` placeholders were used.
                Leave the default value :py:data:`None` if there are no placeholders in the query.
            is_safe: set to :py:data:`True` if the query can be safely run also in DRY-RUN mode. By default all queries
                are considered unsafe. If :py:data:`False` the query will not be run in DRY-RUN mode and the return
                value will be 0.
            **kwargs: arbitrary arguments that are passed to the :py:class:`spicerack.mysql.MysqlClient.connect`
                method. See its documentation for the available arguments and their default values.

        Returns:
            the number of affected rows.

        Raises:
            spicerack.mysql.MysqlError: if the query returned more than one row.
            pymysql.err.MySQLError: on query errors.

        """
        with self.cursor(**kwargs) as (_connection, cursor):
            if self.host.dry_run and not is_safe:
                effective_query = cursor.mogrify(query, query_parameters)
                logger.info("Would have executed on host %s the query: %s", self, effective_query)
                return 0

            num_rows = cursor.execute(query, query_parameters)
            self.check_warnings(cursor)
            return num_rows

    def fetch_one_row(
        self,
        query: str,
        query_parameters: Union[None, tuple, list, dict] = None,
        **kwargs: Any,
    ) -> dict:
        """Execute the given query and returns one row. It sets the connection as read only unless forced explicitly.

        Caution:
            DRY-RUN and read-only support is limited to DML sql operations. By default all queries are considered
            read-only due to the nature of the method (retrieve one row).

        Note:
            If any warning is issued by the database they will be logged and the user prompted what to do.

        Examples:
            ::

                >>> query = "SELECT * FROM mytable WHERE a = %(a)s and b = %(b)s"
                >>> params = {"a": "value", "b": 10}
                >>> row = instance.fetch_one_row(query, params, database="mydb")
                >>> row
                {'a': 'value', 'b': 10, 'c': 1}
                >>> row = instance.fetch_one_row("SELECT 1 WHERE 2 > 1", database="mydb")
                >>> row
                {}

        Arguments:
            query: the query to execute, with eventual placeholders (``%s`` or ``%(name)s``).
            query_parameters: the query parameters to inject into the query, a :py:class:`tuple` or :py:class:`list` in
                case ``%s`` placeholders were used or a :py:class:`dict` in case ``%(name)s`` placeholders were used.
                Leave the default value :py:data:`None` if there are no placeholders in the query.
            **kwargs: arbitrary arguments that are passed to the :py:class:`spicerack.mysql.MysqlClient.connect`
                method. See its documentation for the available arguments and their default values.

        Returns:
            the fetched row or :py:data:`None` if the query returned no rows.

        Raises:
            spicerack.mysql.MysqlError: if the query returned more than one row.
            pymysql.err.MySQLError: on query errors.

        """
        kwargs["read_only"] = kwargs.get("read_only", True)  # Set RO to true unless explicitly specified
        with self.cursor(**kwargs) as (_connection, cursor):
            num_rows = cursor.execute(query, query_parameters)
            if num_rows != 1:
                self.check_warnings(cursor)
                if num_rows == 0:
                    return {}

                raise MysqlError(f"Expected query to return zero or one row, got {num_rows} instead.")

            row = cursor.fetchone()
            self.check_warnings(cursor)

            return row

    def run_query(self, query: str, database: str = "", **kwargs: Any) -> Any:
        """Execute the query via Remote.

        Arguments:
            query: the mysql query to be executed. Double quotes must be already escaped.
            database: the optional database to use for the query execution.
            **kwargs: any additional argument is passed to :py:meth:`spicerack.remote.RemoteHosts.run_sync`. By default
                the ``print_progress_bars`` and ``print_output`` arguments are set to :py:data:`False`.

        Returns:
            The result of the remote command execution.

        Raises:
            spicerack.remote.RemoteExecutionError: if the query execution via SSH returns a non-zero exit code.

        """
        command = f'{self._mysql_bin} --socket {self._sock} --batch --execute "{query}" {database}'.strip()
        kwargs.setdefault("print_progress_bars", False)
        kwargs.setdefault("print_output", False)
        try:
            return self.host.run_sync(command, **kwargs)
        except RemoteExecutionError as e:
            raise MysqlError(f"Failed to run '{query}' on {self.host}") from e

    def run_vertical_query(self, query: str, database: str = "", **kwargs: Any) -> list[dict[str, str]]:
        r"""Run a query with vertical output (terminating it with ``\G``) and parse its results.

        The ``\G`` at the end of the query is automatically added.
        Each returned row is converted to a dictionary with keys that are the column names and values that are
        the column values.

        Warning:
            The parsing of the output of queries from the CLI, even with vertical format (``\G``), is a very brittle
            operation that could fail or have misleading data, for example if any of the values queried are
            multi-lines. This could potentially happen also with a ``SHOW SLAVE STATUS`` query if the replication is
            broken and the last error contains a newline.

        Arguments:
            According to :py:meth:`spicerack.mysql.Instance.run_query`.

        Returns:
            the parsed query as a list of dictionaries, one per returned row.

        """
        response = list(self.run_query(rf"{query}\G", database, **kwargs))
        if not response:
            return []

        rows = []
        current: dict[str, str] = {}
        for line in response[0][1].message().decode("utf8").splitlines():
            if line.startswith("*" * 20) and f" row {'*' * 20}" in line:  # row separator
                if current:
                    rows.append(current)
                current = {}
                continue

            try:
                key, value = line.split(": ", 1)
            except ValueError:
                logger.error("Failed to parse into key/value for query '%s' this line: %s", query, line)
                continue

            current[key.lstrip()] = value

        if current:
            rows.append(current)

        return rows

    def is_running(self) -> bool:
        """Check if the systemd service for the instance is active and running.

        Returns:
            True if the service is active and running, False otherwise.

        """
        command = f"/usr/bin/systemctl show -p ActiveState -p SubState {self._service}"

        result = self.host.run_sync(command, is_safe=True)
        output = list(result)[0][1].message().decode("utf-8").strip()

        return "ActiveState=active" in output and "SubState=running" in output

    def stop_slave(self) -> None:
        """Stops mariadb replication."""
        self.execute("STOP SLAVE")

    def start_slave(self) -> None:
        """Starts mariadb replication and sleeps for 1 second afterwards."""
        self.execute("START SLAVE")
        sleep(1)

    def show_slave_status(self) -> dict:
        """Returns the output of show slave status formatted as a dict.

        Returns:
            the current slave status for the instance.

        """
        query = "SHOW SLAVE STATUS"
        with self.cursor() as (_connection, cursor):
            num_rows = cursor.execute(query)
            if not num_rows:
                raise MysqlError(f"{query} seems to have been executed on a master.")

            if num_rows > 1:
                raise NotImplementedError(f"Multisource setup are not implemented. Got {num_rows} rows.")

            return cursor.fetchone()

    def show_master_status(self) -> dict:
        """Returns the output of show master status formatted as a dict.

        Returns:
            the current master status for the instance.

        """
        query = "SHOW MASTER STATUS"
        status = self.fetch_one_row(query)
        if not status:
            raise MysqlError(f"{query} seems to have been executed on a host with binlog disabled.")

        return status

    def set_master_use_gtid(self, setting: MasterUseGTID) -> None:
        """Runs MASTER_USE_GTID with the given value."""
        if not isinstance(setting, MasterUseGTID):
            raise MysqlError(f"Only instances of MasterUseGTID are accepted, got: {type(setting)}")

        # Not using placeholder replacements ad MariaDB requires it to be a syntax word and raises if it's quoted
        self.execute(f"CHANGE MASTER TO MASTER_USE_GTID={setting.value}")

    def stop_mysql(self) -> Iterator[tuple[NodeSet, MsgTreeElem]]:
        """Stops mariadb service.

        Returns:
            The results of the remote status command.

        """
        return self.host.run_sync(f"/usr/bin/systemctl stop {self._service}")

    def status_mysql(self) -> Iterator[tuple[NodeSet, MsgTreeElem]]:
        """Stops mariadb service.

        Returns:
            The results of the remote status command. It does not raise on exit codes different from zero.

        """
        return self.host.run_sync(Command(f"/usr/bin/systemctl status {self._service}", ok_codes=[]), is_safe=True)

    def start_mysql(self) -> Iterator[tuple[NodeSet, MsgTreeElem]]:
        """Starts mariadb service.

        Returns:
            The results of the remote start command.

        """
        return self.host.run_sync(f"/usr/bin/systemctl start {self._service}", print_output=True)

    def restart_mysql(self) -> Iterator[tuple[NodeSet, MsgTreeElem]]:
        """Restarts mariadb service.

        Returns:
            The results of the remote restart command.

        """
        return self.host.run_sync(f"/usr/bin/systemctl restart {self._service}")

    def clean_data_dir(self, *, skip_confirmation: bool = False) -> None:
        """Removes everything contained in the data directory.

        Arguments:
            skip_confirmation: execute the operation without any user confirmation.

        """
        command = f"/usr/bin/rm -rf {self._data_dir}"
        if not skip_confirmation:
            ask_confirmation(f"ATTENTION: destructive action for {self.host}: {command}. Are you sure to proceed?")

        self.host.run_sync(command)

    def upgrade(self) -> Iterator[tuple[NodeSet, MsgTreeElem]]:
        """Runs the relevant mysql_upgrade command to upgrade the instance content.

        Returns:
            The results of the remote upgrade command.

        """
        command = f"$(readlink -f /usr/local/bin/mysql_upgrade) --socket {self._sock} --force"
        return self.host.run_sync(command, print_output=True)

    def get_replication_info(self) -> ReplicationInfo:
        """Get the replication information suitable to set a new node's replication.

        Returns:
            The replication information object, useful to setup a new instance's replication to resume from the same
            position.

        """
        replication_status = self.show_slave_status()
        info = ReplicationInfo(
            primary=replication_status.get("Master_Host", ""),
            binlog=replication_status.get("Master_Log_File", ""),
            position=int(replication_status.get("Exec_Master_Log_Pos", -1)),
            port=int(replication_status.get("Master_Port", -1)),
        )
        if not (info.primary and info.binlog and info.position > -1 and info.port > -1):
            raise MysqlError(f"Could not find the replication position: {info}")

        logger.debug("Replication info for %s: %s", self.host, info)
        return info

    @property
    def data_dir(self) -> str:
        """Get the data directory of this instance.

        Returns:
            the data directory path for mariadb for this specific instance.

        """
        return self._data_dir

    @property
    def socket(self) -> str:
        """Getter for the socket path of the instance.

        Returns:
            the instance specific socket path to use.

        """
        return self._sock

    @property
    def primary(self) -> str:
        """Retrieves the replication source of this cluster.

        Raises:
            spicerack.mysql.MysqlError: if unable to find the master host of the current instance.

        """
        if not self._primary:
            try:
                self._primary = self.show_slave_status()["Master_Host"]
            except (KeyError, MysqlError) as e:
                raise MysqlError("Unable to retrieve master host") from e

        return self._primary

    def prep_src_for_cloning(self) -> ReplicationInfo:
        """Helper that prepares source instance to be cloned.

        Returns:
            The replication information object, useful to setup a new instance's replication to resume from the same
            position.

        """
        self.stop_slave()
        replication_info = self.get_replication_info()
        self.stop_mysql()
        return replication_info

    def set_replication_parameters(self, *, replication_info: ReplicationInfo, user: str, password: str) -> None:
        """Sets the replication parameters for the MySQL instance."""
        self.execute(
            (
                "CHANGE MASTER TO master_host=%(primary)s, "
                "master_port=%(port)s, "
                "master_ssl=1, "
                "master_log_file=%(binlog)s, "
                "master_log_pos=%(position)s, "
                "master_user=%(user)s, "
                "master_password=%(password)s"
            ),
            {
                "primary": replication_info.primary,
                "port": replication_info.port,
                "binlog": replication_info.binlog,
                "position": replication_info.position,
                "user": user,
                "password": password,
            },
        )

    def post_clone_reset_with_slave_stopped(self) -> None:
        """Prepares the MySQL instance for the first pooling operation."""
        self.host.run_sync(
            f"chown -R mysql:mysql {self._data_dir}",
            '/usr/bin/systemctl set-environment MYSQLD_OPTS="--skip-slave-start"',
        )
        self.start_mysql()
        self.stop_slave()
        self.execute("RESET SLAVE ALL")

    def resume_replication(self) -> None:
        """Resumes replication on the source MySQL instance."""
        self.host.run_sync('/usr/bin/systemctl set-environment MYSQLD_OPTS="--skip-slave-start"')
        self.start_mysql()
        self.upgrade()
        self.restart_mysql()
        self.start_slave()

    @retry(
        tries=480,  # We allow up to 8 hours for replication lag to catch up
        delay=timedelta(seconds=60),
        backoff_mode="constant",
        exceptions=(MysqlReplagError,),
    )
    def wait_for_replication(self, threshold: Union[float, Decimal] = Decimal("1.0")) -> None:
        """Waits for replication to catch up.

        Arguments:
            threshold: the replication lag threshold in seconds under which the replication is considered in sync.

        Raises:
            spicerack.mysql.MysqlReplagError: if the replication lag is still too high after all the
                retries.

        """
        replag = self.replication_lag()
        if replag > threshold:
            raise MysqlReplagError(f"Replication lag higher than the threshold ({threshold}s): {replag}s")

    def replication_lag(self) -> Decimal:
        """Retrieves the current replication lag.

        Returns:
            The replication lag in seconds.

        Raises:
            spicerack.mysql.MysqlError: if no lag information is present or unable to parse the it.

        """
        query = (
            "SELECT greatest(0, TIMESTAMPDIFF(MICROSECOND, max(ts), UTC_TIMESTAMP(6)) - 500000)/1000000 AS lag "
            "FROM heartbeat ORDER BY ts LIMIT 1"
        )
        row = self.fetch_one_row(query, database="heartbeat")
        if not row:
            raise MysqlError("The replication lag query returned no data")

        if "lag" not in row or row["lag"] is None:
            raise MysqlError(f"Unable to get lag information from: {row}")

        return row["lag"]


class MysqlRemoteHosts(RemoteHostsAdapter):
    """Custom RemoteHosts class for executing MySQL queries."""

    def __iter__(self) -> Iterator["MysqlRemoteHosts"]:
        """Iterate over all remote hosts in this instance.

        Yields:
            spicerack.mysql.MysqlRemoteHosts: an new instance for each host.

        """
        yield from self.split(len(self))

    def split(self, n_slices: int) -> Iterator["MysqlRemoteHosts"]:
        """Split the current MySQL remote hosts instance into ``n_slices`` instances.

        Arguments:
            n_slices: the number of slices to slice the MySQL remote hosts into.

        Yields:
            spicerack.mysql.MysqlRemoteHosts: the instances for the subset of nodes.

        """
        for remote_hosts in self.remote_hosts.split(n_slices):
            yield MysqlRemoteHosts(remote_hosts)

    # TODO merge this method with Instance.run_query()
    def run_query(self, query: str, database: str = "", **kwargs: Any) -> Iterator[tuple[NodeSet, MsgTreeElem]]:
        """Execute the query via Remote.

        Arguments:
            query: the mysql query to be executed. Double quotes must be already escaped.
            database: an optional MySQL database to connect to before executing the query.
            **kwargs: any additional argument is passed to :py:meth:`spicerack.remote.RemoteHosts.run_sync`. By default
                the ``print_progress_bars`` and ``print_output`` arguments are set to :py:data:`False`.

        Raises:
            spicerack.remote.RemoteExecutionError: if the Cumin execution returns a non-zero exit code.

        """
        command = f'/usr/local/bin/mysql --skip-ssl --skip-column-names --batch -e "{query}" {database}'.strip()
        kwargs.setdefault("print_progress_bars", False)
        kwargs.setdefault("print_output", False)
        return self._remote_hosts.run_sync(command, **kwargs)

    def list_hosts_instances(self, *, group: bool = False) -> list[Instance]:
        """List MariaDB instances on the host.

        Arguments:
            group: not yet implemented feature to allow parallelization.

        Raises:
            spicerack.remote.RemoteExecutionError: if the Cumin execution returns a non-zero exit code.
            NotImplementedError: if the replag is not fully caught on.

        """
        if len(self._remote_hosts) != 1:
            raise NotImplementedError("Only single host are supported at this time.")

        if group:
            # TODO see this comment:
            # https://gerrit.wikimedia.org/r/c/operations/software/spicerack/+/1005531/comment/7af929a5_6d6184d4/
            # we could use this method to parallelize stuff on instances as well.
            raise NotImplementedError("Grouping and parallelization are not supported at this time.")

        return self._list_host_instances()

    def _list_host_instances(self) -> list[Instance]:
        """List MariaDB instances on the host.

        Raises:
            spicerack.remote.RemoteExecutionError: if the Cumin execution returns a non-zero exit code.

        """
        instances: list[Instance] = []

        # First check for multi-instance configs
        conf_files = list(
            self._remote_hosts.run_sync(
                # Exclude non-readable directories to avoid spurious error messages, select only .cnf files and
                # print their name without the path.
                r'/usr/bin/find /etc ! -readable -prune -o -path "/etc/mysql/mysqld.conf.d/*.cnf" -printf "%f\n"',
                is_safe=True,
                print_output=False,
                print_progress_bars=False,
            )
        )
        if conf_files:  # Multi-instance
            for conf_file in conf_files[0][1].message().decode().splitlines():
                instances.append(Instance(self._remote_hosts, name=conf_file[:-4]))  # Remove .cnf extension
        else:  # Check for single instance
            try:
                self._remote_hosts.run_sync(
                    r"/usr/bin/grep -q '^\[mysqld\]$' '/etc/my.cnf'",
                    is_safe=True,
                    print_output=False,
                    print_progress_bars=False,
                )
                instances.append(Instance(self._remote_hosts))
            except RemoteExecutionError:  # No my.cnf or no mysqld section found - no instances present
                pass

        return instances


class Mysql:
    """Class to manage MySQL servers."""

    # FIXME this query could be replaced by the one in _get_replication() as it's the one used in monitoring
    heartbeat_query: str = (
        "SELECT ts FROM heartbeat.heartbeat WHERE datacenter = '{dc}' and shard = '{section}' "
        "ORDER BY ts DESC LIMIT 1"
    )
    """Query pattern to check the heartbeat for a given datacenter and section."""

    def __init__(self, remote: Remote, dry_run: bool = True) -> None:
        """Initialize the instance.

        Arguments:
            remote: the Remote instance.
            dry_run: whether this is a DRY-RUN.

        """
        self._remote = remote
        self._dry_run = dry_run

    def get_dbs(self, query: str) -> MysqlRemoteHosts:
        """Get a MysqlRemoteHosts instance for the matching hosts.

        Arguments:
            query: the Remote query to use to fetch the DB hosts.

        """
        return MysqlRemoteHosts(self._remote.query(query))

    def get_core_dbs(
        self,
        *,
        datacenter: Optional[str] = None,
        section: Optional[str] = None,
        replication_role: Optional[str] = None,
        excludes: tuple[str, ...] = (),
    ) -> MysqlRemoteHosts:
        """Get an instance to operated on the core databases matching the parameters.

        Arguments:
            datacenter: the name of the datacenter to filter for, accepted values are those specified in
                :py:data:`spicerack.constants.CORE_DATACENTERS`.
            replication_role: the repication role to filter for, accepted values are those specified in
                :py:data:`spicerack.mysql.REPLICATION_ROLES`.
            section: a specific section to filter for, accepted values are those specified in
                :py:data:`spicerack.mysql.CORE_SECTIONS`.
            excludes: sections to exclude from getting.

        Raises:
            spicerack.mysql.MysqlError: on invalid data or unexpected matching hosts.

        """
        query_parts = ["A:db-core"]
        dc_multipler = len(CORE_DATACENTERS)
        section_multiplier = len(CORE_SECTIONS)

        if datacenter is not None:
            dc_multipler = 1
            if datacenter not in CORE_DATACENTERS:
                raise MysqlError(f"Got invalid datacenter {datacenter}, accepted values are: {CORE_DATACENTERS}")

            query_parts.append("A:" + datacenter)

        for exclude in excludes:
            if exclude not in CORE_SECTIONS:
                raise MysqlError(f"Got invalid excludes {exclude}, accepted values are: {CORE_SECTIONS}")
            section_multiplier -= 1
            query_parts.append(f"not A:db-section-{exclude}")

        if section is not None:
            section_multiplier = 1
            if section not in CORE_SECTIONS:
                raise MysqlError(f"Got invalid section {section}, accepted values are: {CORE_SECTIONS}")

            query_parts.append(f"A:db-section-{section}")

        if replication_role is not None:
            if replication_role not in REPLICATION_ROLES:
                raise MysqlError(
                    f"Got invalid replication_role {replication_role}, accepted values are: {REPLICATION_ROLES}"
                )

            query_parts.append(f"A:db-role-{replication_role}")

        mysql_hosts = MysqlRemoteHosts(self._remote.query(" and ".join(query_parts)))

        # Sanity check of matched hosts in case of master selection
        if replication_role == "master" and len(mysql_hosts) != dc_multipler * section_multiplier:
            raise MysqlError(f"Matched {len(mysql_hosts)} masters, expected {dc_multipler * section_multiplier}")

        return mysql_hosts

    def set_core_masters_readonly(self, datacenter: str) -> None:
        """Set the core masters in read-only.

        Arguments:
            datacenter: the name of the datacenter to filter for.

        Raises:
            spicerack.remote.RemoteExecutionError: on Remote failures.
            spicerack.mysql.MysqlError: on failing to verify the modified value.

        """
        logger.debug("Setting core DB masters in %s to be read-only", datacenter)
        target = self.get_core_dbs(datacenter=datacenter, replication_role="master")
        target.run_query("SET GLOBAL read_only=1")
        self.verify_core_masters_readonly(datacenter, True)

    def set_core_masters_readwrite(self, datacenter: str) -> None:
        """Set the core masters in read-write.

        Arguments:
            datacenter: the name of the datacenter to filter for.

        Raises:
            spicerack.remote.RemoteExecutionError: on Remote failures.
            spicerack.mysql.MysqlError: on failing to verify the modified value.

        """
        logger.debug("Setting core DB masters in %s to be read-write", datacenter)
        target = self.get_core_dbs(datacenter=datacenter, replication_role="master")
        target.run_query("SET GLOBAL read_only=0")
        self.verify_core_masters_readonly(datacenter, False)

    def verify_core_masters_readonly(self, datacenter: str, is_read_only: bool) -> None:
        """Verify that the core masters are in read-only or read-write mode.

        Arguments:
            datacenter: the name of the datacenter to filter for.
            is_read_only: whether the read-only mode should be set or not.

        Raises:
            spicerack.mysql.MysqlError: on failure.

        """
        logger.debug(
            "Verifying core DB masters in %s have read-only=%d",
            datacenter,
            is_read_only,
        )
        target = self.get_core_dbs(datacenter=datacenter, replication_role="master")
        expected = str(int(is_read_only))  # Convert it to the returned value from MySQL: 1 or 0.
        failed = False

        for nodeset, output in target.run_query("SELECT @@global.read_only", is_safe=True):
            response = output.message().decode()
            if response != expected:
                logger.error(
                    "Expected output to be '%s', got '%s' for hosts %s",
                    expected,
                    response,
                    str(nodeset),
                )
                failed = True

        if failed and not self._dry_run:
            raise MysqlError(f"Verification failed that core DB masters in {datacenter} have read-only={is_read_only}")

    def check_core_masters_in_sync(self, dc_from: str, dc_to: str) -> None:
        """Check that all core masters in dc_to are in sync with the core masters in dc_from.

        Arguments:
            dc_from: the name of the datacenter from where to get the master positions.
            dc_to: the name of the datacenter where to check that they are in sync.

        Raises:
            spicerack.remote.RemoteExecutionError: on failure.

        """
        logger.debug("Waiting for the core DB masters in %s to catch up", dc_to)
        heartbeats = self.get_core_masters_heartbeats(dc_from, dc_from)
        self.check_core_masters_heartbeats(dc_to, dc_from, heartbeats)

    def get_core_masters_heartbeats(self, datacenter: str, heartbeat_dc: str) -> dict[str, datetime]:
        """Get the current heartbeat values from core DB masters in DC for a given heartbeat DC.

        Arguments:
            datacenter: the name of the datacenter from where to get the heartbeat values.
            heartbeat_dc: the name of the datacenter for which to filter the heartbeat query.

        Returns:
            A dictionary with the section name :py:class:`str` as keys and their heartbeat
            :py:class:`datetime.datetime` as values. For example::

                {'s1': datetime.datetime(2018, 1, 2, 11, 22, 33, 123456)}

        Raises:
            spicerack.mysql.MysqlError: on failure to gather the heartbeat or convert it into a datetime.

        """
        heartbeats = {}
        for section in CORE_SECTIONS:
            core_dbs = self.get_core_dbs(datacenter=datacenter, section=section, replication_role="master")
            heartbeats[section] = Mysql._get_heartbeat(core_dbs, section, heartbeat_dc)

        return heartbeats

    def check_core_masters_heartbeats(
        self, datacenter: str, heartbeat_dc: str, heartbeats: dict[str, datetime]
    ) -> None:
        """Check the current heartbeat values in the core DB masters in DC are in sync with the provided heartbeats.

        Arguments:
            datacenter: the name of the datacenter from where to get the heartbeat values.
            heartbeat_dc: the name of the datacenter for which to filter the heartbeat query.
            heartbeats: a dictionary with the section name :py:class:`str` as keys and heartbeat
                :py:class:`datetime.datetime` for each core section as values.

        Raises:
            spicerack.mysql.MysqlError: on failure to gather the heartbeat or convert it into a datetime.

        """
        for section, heartbeat in heartbeats.items():
            self._check_core_master_in_sync(datacenter, heartbeat_dc, section, heartbeat)

    @retry(exceptions=(MysqlError,))
    def _check_core_master_in_sync(
        self,
        datacenter: str,
        heartbeat_dc: str,
        section: str,
        parent_heartbeat: datetime,
    ) -> None:
        """Check and retry that the heartbeat value in a core DB master in DC is in sync with the provided heartbeat.

        Arguments:
            datacenter: the name of the datacenter from where to get the heartbeat value.
            heartbeat_dc: the name of the datacenter for which to filter the heartbeat query.
            section: the section name from where to get the heartbeat value and filter the heartbeat query.
            master_heartbeat: the reference heartbeat from the parent master to use to verify this master is in sync
                with it.

        Raises:
            spicerack.mysql.MysqlError: on failure to gather the heartbeat or convert it into a datetime
                or not yet in sync.

        """
        core_dbs = self.get_core_dbs(datacenter=datacenter, section=section, replication_role="master")
        local_heartbeat = Mysql._get_heartbeat(core_dbs, section, heartbeat_dc)

        # The check requires that local_heartbeat is stricly greater than parent_heartbeat because heartbeat writes also
        # when the DB is in read-only mode and has a granularity of 1s (as of 2018-09), meaning that an event could have
        # been written after the last heartbeat but before the DB was set in read-only mode and that event could not
        # have been replicated, hence checking the next heartbeat to ensure they are in sync.
        if local_heartbeat <= parent_heartbeat:
            delta = (local_heartbeat - parent_heartbeat).total_seconds()
            raise MysqlError(
                f"Heartbeat from master {core_dbs} for section {section} not yet in sync: "
                f"{local_heartbeat} <= {parent_heartbeat} (delta={delta})"
            )

    @staticmethod
    def _get_heartbeat(mysql_hosts: MysqlRemoteHosts, section: str, heartbeat_dc: str) -> datetime:
        """Get the heartbeat from the remote host for a given DC.

        Arguments:
            mysql_hosts: the instance for the target DB to query.
            section: the DB section for which to get the heartbeat.
            heartbeat_dc: the name of the datacenter for which to filter the heartbeat query.

        Raises:
            spicerack.mysql.MysqlError: on failure to gather the heartbeat or convert it into a datetime.

        """
        query = Mysql.heartbeat_query.format(dc=heartbeat_dc, section=section)

        for _, output in mysql_hosts.run_query(query, is_safe=True):
            try:
                heartbeat_str = output.message().decode()
                heartbeat = datetime.strptime(heartbeat_str, "%Y-%m-%dT%H:%M:%S.%f")
                break
            except (TypeError, ValueError) as e:
                raise MysqlError(f"Unable to convert heartbeat '{heartbeat_str}' into datetime") from e
        else:
            raise MysqlError(f"Unable to get heartbeat from master {mysql_hosts} for section {section}")

        return heartbeat



"""Apiclient module tests."""

import logging
import re

import pytest
from requests.exceptions import ConnectTimeout
from requests_mock.adapter import ANY
from wmflib.requests import http_session

from spicerack import apiclient


class TestAPIClient:
    """Tests for the APIClient class."""

    @pytest.fixture(autouse=True)
    def setup_method(self, requests_mock):
        """Initialize the test instance."""
        # pylint: disable=attribute-defined-outside-init
        self.requests_mock = requests_mock
        self.base_url = "https://api.example.org/v1"
        self.endpoint = f"{self.base_url}/endpoint"
        self.http_session = http_session("APIClient Tests")
        self.apiclient = apiclient.APIClient(self.base_url, self.http_session, dry_run=False)
        self.apiclient_dry_run = apiclient.APIClient(self.base_url, self.http_session)

    def test_http_session(self):
        """It should return the current HTTP session object to allow the user to fine-tune it."""
        assert self.apiclient.http_session is self.http_session
        self.apiclient.http_session.headers.update({"Accept": "application/json"})
        self.requests_mock.get(self.endpoint, json={"key": "value"})
        response = self.apiclient.request("get", "/endpoint")
        assert response.request.headers["Accept"] == "application/json"

    @pytest.mark.parametrize("method", ("get", "head", "options"))
    def test_request_dry_run_ro(self, method):
        """It should perform any RO request and return the actual response also in dry_run mode."""
        kwargs = {}
        if method == "get":
            kwargs["json"] = {"key": "value"}

        self.requests_mock.request(method, self.endpoint, **kwargs)
        response = self.apiclient_dry_run.request(method, "/endpoint")
        assert response.status_code == 200
        if method == "get":
            assert response.json() == {"key": "value"}

    @pytest.mark.parametrize("method", ("connect", "delete", "patch", "post", "put", "trace"))
    def test_request_dry_run_rw(self, method):
        """It should not perform any RW request and return a dummy successful response in dry-run mode."""
        self.requests_mock.request(ANY, ANY, status_code=500)
        response = self.apiclient_dry_run.request(method, "/endpoint")
        assert response.status_code == 200
        assert not response.text

    def test_request_dry_run_fail(self, caplog):
        """If the request fails in dry-run mode, it should return a dummy successful response."""
        self.requests_mock.get(self.endpoint, exc=ConnectTimeout)
        with caplog.at_level(logging.ERROR):
            response = self.apiclient_dry_run.request("get", "/endpoint")

        assert response.status_code == 200
        assert f"Failed to perform GET request to {self.endpoint}" in caplog.text

    def test_request_ok(self):
        """It should perform the provided request and return it."""
        self.requests_mock.get(self.endpoint, json={"key": "value"})
        response = self.apiclient.request("get", "/endpoint")
        assert response.json() == {"key": "value"}
        assert response.status_code == 200

    def test_request_response_wrong_status_code(self):
        """It should raise an APIClientResponseError if the request returns an error status code."""
        self.requests_mock.post(self.endpoint, json={"error_code": "01", "message": "error"}, status_code=405)
        with pytest.raises(
            apiclient.APIClientResponseError, match=re.escape(f"POST {self.endpoint} returned HTTP 405")
        ) as e:
            self.apiclient.request("post", "/endpoint", json={"key": "value"})
            assert e.response.status_code == 405
            assert e.response.json()["error_code"] == "01"

    def test_request_response_raises(self):
        """It should raise an APIClientError if the request failes to be performed."""
        self.requests_mock.get(self.endpoint, exc=ConnectTimeout)
        with pytest.raises(
            apiclient.APIClientError, match=re.escape(f"Failed to perform GET request to {self.endpoint}")
        ):
            self.apiclient.request("get", "/endpoint")

    def test_request_invalid_uri(self):
        """It should raise an TestAPIClient if the URI is invalid."""
        with pytest.raises(apiclient.APIClientError, match=re.escape("Invalid uri 'endpoint', it must start with a /")):
            self.apiclient.request("get", "endpoint")



"""Run something and abort successfully in init."""

import argparse

from spicerack.cookbook import CookbookBase, CookbookInitSuccess, CookbookRunnerBase


class AbortInitCookbook(CookbookBase):
    """A cookbook that aborts successfully in its runner's init."""

    owner_team = "team1"

    def argument_parser(self):
        """As defined by the parent class."""
        parser = argparse.ArgumentParser("Argparse")
        parser.add_argument("--message", help="A message to log.")
        return parser

    def get_runner(self, args):
        """As required by the parent class."""
        return ExampleRunner(args, self.spicerack)


class ExampleRunner(CookbookRunnerBase):
    """As required by spicerack API."""

    def __init__(self, args, _spicerack):
        """Initialize the instance."""
        if args.message:
            raise CookbookInitSuccess(args.message)
        raise CookbookInitSuccess()

    def run(self):
        """As required by the parent class."""
        raise RuntimeError()  # This should never be executed



"""Group3 argparse_ok."""

import argparse

__owner_team__ = "team1"


def argument_parser():
    """As required by spicerack._cookbook."""
    return argparse.ArgumentParser("group3.argparse_ok")


def run(_args, _spicerack):
    """As required by spicerack._cookbook."""



"""Interactive module tests."""

import logging
import uuid
from pathlib import Path
from unittest import mock

import pytest
from wmflib import interactive

from spicerack import _log as log

GENERIC_LOG_RECORD = logging.LogRecord("module", logging.DEBUG, "/source/file.py", 1, "message", [], None)
CUMIN_LOG_RECORD = logging.LogRecord("cumin.module", logging.DEBUG, "/cumin/source/file.py", 1, "message", [], None)
logger = logging.getLogger(__name__)


def reset_logging_module():
    """Reset the logging module removing all handlers and filters."""
    for log_logger in (log.root_logger, log.irc_logger, log.sal_logger):
        list(map(log_logger.removeHandler, log_logger.handlers))
        list(map(log_logger.removeFilter, log_logger.filters))

    for handler in log.notify_logger.handlers:
        if not isinstance(handler, logging.NullHandler):
            log.notify_logger.removeHandler(handler)


def test_cumin_filter_pass():
    """The FilterOutCumin filter() method should let a normal log record pass."""
    log_filter = log.FilterOutCumin()
    ret = log_filter.filter(GENERIC_LOG_RECORD)
    assert ret == 1


def test_cumin_filter_blocks_cumin():
    """The FilterOutCumin filter() method should block a Cumin's log record."""
    log_filter = log.FilterOutCumin()
    ret = log_filter.filter(CUMIN_LOG_RECORD)
    assert ret == 0


class TestLogModule:
    """Test class for the _log module."""

    def setup_method(self):
        """Initialize the test method."""
        # pylint: disable=attribute-defined-outside-init
        self.message = str(uuid.uuid4())

    def teardown_method(self):
        """Reset python's logging module."""
        reset_logging_module()

    def _assert_match_in_tmpdir(self, match, tmp_dir, negate=False):
        """Given a match string, assert that it's present in all files in tmp_dir."""
        tmp_dir = Path(tmp_dir)  # Newer versions pass a LocalPath, older a string.
        for logfile in tmp_dir.iterdir():
            content = (tmp_dir / logfile).read_text()
            if negate:
                assert match not in content
            else:
                assert match in content

    def test_setup_logging_no_irc(self, tmpdir, caplog):
        """Calling setup_logging() should setup all the handlers of the root logger."""
        log.setup_logging(Path(tmpdir.strpath), "task", "user")
        logger.info(self.message)
        assert self.message in caplog.text
        self._assert_match_in_tmpdir(self.message, tmpdir.strpath)

    @mock.patch("wmflib.irc.socket")
    def test_setup_logging_with_irc(self, mocked_socket, tmpdir, caplog):
        """Calling setup_logging() with host and port should also setup the IRC loggers."""
        log.setup_logging(Path(tmpdir.strpath), "task", "user", dry_run=False, host="host", port=123)
        irc_message = self.message
        sal_message = str(uuid.uuid4())
        log.irc_logger.info(irc_message)
        log.sal_logger.info(sal_message)

        assert mock.call.socket().connect(("host", 123)) in mocked_socket.mock_calls
        assert irc_message in caplog.text
        assert sal_message in caplog.text
        self._assert_match_in_tmpdir(irc_message, tmpdir.strpath)
        self._assert_match_in_tmpdir(sal_message, tmpdir.strpath)

    def test_setup_logging_dry_run(self, capsys, tmpdir, caplog):
        """Calling setup_logging() when in dry run mode should setup all the handlers and the stdout in DEBUG."""
        log.setup_logging(Path(tmpdir.strpath), "task", "user")
        logger.info(self.message)

        _, err = capsys.readouterr()
        assert self.message in err
        assert "DRY-RUN" in err
        assert self.message in caplog.text
        self._assert_match_in_tmpdir(self.message, tmpdir.strpath)
        self._assert_match_in_tmpdir("DRY-RUN", tmpdir.strpath)

    @pytest.mark.parametrize("notify_logger_enabled", (True, False))
    @mock.patch("wmflib.interactive.NOTIFY_AFTER_SECONDS", 0.0)
    @mock.patch("builtins.input", return_value="go")
    @mock.patch("wmflib.interactive.sys.stdout.isatty", return_value=True)
    @mock.patch("wmflib.irc.socket.socket")
    def test_setup_logging_notify_logger_on(  # pylint: disable=too-many-positional-arguments,too-many-arguments
        self, mocked_socket, mocked_isatty, mocked_input, notify_logger_enabled, capsys, tmpdir, caplog
    ):
        """It should setup the wmflib's notify_logger based on the related parameter."""
        log.setup_logging(
            Path(tmpdir.strpath),
            "task",
            "user",
            dry_run=False,
            host="host",
            port=123,
            notify_logger_enabled=notify_logger_enabled,
        )
        awaiting = "is awaiting input"

        interactive.ask_confirmation(self.message)

        out, err = capsys.readouterr()
        assert self.message in out
        assert awaiting not in out
        assert self.message not in err
        assert awaiting not in err
        assert self.message not in caplog.text
        assert awaiting not in caplog.text
        self._assert_match_in_tmpdir(self.message, tmpdir.strpath, negate=True)
        mocked_isatty.assert_called_once_with()
        mocked_input.assert_called_once_with("> ")

        if notify_logger_enabled:
            mocked_socket.assert_called()
            sendall = mocked_socket.return_value.sendall
            sendall.assert_called_once()
            assert awaiting.encode() in sendall.call_args.args[0]
        else:
            mocked_socket.assert_not_called()

    @pytest.mark.parametrize("skip_start_sal", (True, False))
    @mock.patch("wmflib.irc.socket.socket")
    def test_log_task_start(self, mocked_socket, skip_start_sal, capsys, tmpdir, caplog):
        """Calling log_task_start() should log a START message for the task to both loggers based on skip_start_sal."""
        log.setup_logging(Path(tmpdir.strpath), "task", "user", dry_run=False, host="host", port=123)
        log.log_task_start(skip_start_sal=skip_start_sal, message=self.message)

        logged_message = f"START - {self.message}"
        _, err = capsys.readouterr()
        assert logged_message in err
        assert logged_message in caplog.text
        self._assert_match_in_tmpdir(logged_message, tmpdir.strpath)
        if skip_start_sal:
            mocked_socket.assert_not_called()
        else:
            sendall = mocked_socket.return_value.sendall
            sendall.assert_called_once()
            assert logged_message.encode() in sendall.call_args.args[0]

    @pytest.mark.parametrize("skip_start_sal", (True, False))
    @mock.patch("wmflib.irc.socket.socket")
    def test_log_task_start_dry_run(self, mocked_socket, skip_start_sal, capsys, tmpdir, caplog):
        """Calling log_task_start() in dry-run mode should not print a START message for the task to the IRC logger."""
        log.setup_logging(Path(tmpdir.strpath), "task", "user", host="host", port=123)
        log.log_task_start(skip_start_sal=skip_start_sal, message=self.message)

        logged_message = f"START - {self.message}"
        _, err = capsys.readouterr()
        assert logged_message in err
        assert logged_message in caplog.text
        self._assert_match_in_tmpdir(logged_message, tmpdir.strpath)
        self._assert_match_in_tmpdir("DRY-RUN", tmpdir.strpath)
        mocked_socket.assert_not_called()

    @pytest.mark.parametrize("skip_start_sal", (True, False))
    @mock.patch("wmflib.irc.socket.socket")
    def test_log_task_end(self, mocked_socket, skip_start_sal, capsys, tmpdir, caplog):
        """Calling log_task_end() should print an END or DONE message for the task."""
        log.setup_logging(Path(tmpdir.strpath), "task", "user", dry_run=False, host="host", port=123)
        log.log_task_end(skip_start_sal=skip_start_sal, status="success", message=self.message)

        if skip_start_sal:
            prefix = "DONE"
        else:
            prefix = "END"

        logged_message = f"{prefix} (success) - {self.message}"
        _, err = capsys.readouterr()
        assert logged_message in err
        assert logged_message in caplog.text
        self._assert_match_in_tmpdir(logged_message, tmpdir.strpath)
        sendall = mocked_socket.return_value.sendall
        sendall.assert_called_once()
        assert logged_message.encode() in sendall.call_args.args[0]



"""Top level class external cookbook."""

from spicerack.cookbook import CookbookBase, CookbookRunnerBase


class ExampleCookbook(CookbookBase):
    """Top level class external cookbook."""

    def get_runner(self, args):
        """As required by the parent class."""
        return ExampleRunner()


class ExampleRunner(CookbookRunnerBase):
    """As required by spicerack._cookbook."""

    def run(self):
        """As required by the parent class."""



"""Constants module."""

WMF_CA_BUNDLE_PATH: str = "/etc/ssl/certs/wmf-ca-certificates.crt"
"""The path to the internal WMF CA certificates bundle, includes the old Puppet CA."""

KEYHOLDER_SOCK: str = "/run/keyholder/proxy.sock"
"""The path to the keyholder agent sock file used for cumin remote commands."""



"""Cookbook module."""

import argparse
import re
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import Optional

from spicerack import Spicerack

ROLLBACK_FAIL_RETCODE: int = 93
"""Reserved exit code: the cookbook had failed and raised an exception while executing the rollback() method."""
CLASS_FAIL_INIT_RETCODE: int = 94
"""Reserved exit code: failed to initialize the cookbook."""
GET_ARGS_PARSER_FAIL_RETCODE: int = 95
"""Reserved exit code: the call to get the argument parser failed."""
PARSE_ARGS_FAIL_RETCODE: int = 96
"""Reserved exit code: a cookbook failed to parse arguments."""
INTERRUPTED_RETCODE: int = 97
"""Reserved exit code: a cookbook execution was interrupted."""
NOT_FOUND_RETCODE: int = 98
"""Reserved exit code: no cookbook was found for the selection."""
EXCEPTION_RETCODE: int = 99
"""Reserved exit code: a cookbook raised an exception while executing."""


class CookbookInitSuccess(Exception):
    """Custom exception class to interrupt the execution before ``run()`` is called in a successful way.

    If a cookbook raises this exception in its runner's ``__init__()`` method, Spicerack will consider the execution
    successful, will not print any stack trace and the exit code will be 0.
    This is useful if the cookbook has some read-only mode where it runs just some checks or gather some data and
    doesn't want to execute anything. Bailing out early in the ``__init__()`` allows also to skip any logging to SAL.
    If the exception is raised with any message that message will be logged with INFO level.
    """


class ArgparseFormatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter):
    """Custom argparse formatter class for cookbooks.

    It can be used as the ``formatter_class`` for the ``ArgumentParser`` instances and it has the capabilities of
    both :py:class:`argparse.ArgumentDefaultsHelpFormatter` and :py:class:`argparse.RawDescriptionHelpFormatter`.
    """


@dataclass(frozen=True)
class LockArgs:
    """A dataclass to represent the arguments to use for the cookbook automatically acquired lock for each run.

    To be used when a cookbook overrides the ``lock_args`` property.

    Arguments:
        suffix: a custom suffix to add to the lock key. The lock key is based on the cookbook's full name and the
            suffix is added to it with a colon separator, for example with suffix ``ro`` the key will become
            ``sre.some.cookbook:ro``.
        concurrency: how many parallel runs of the cookbook with the same lock arguments can be run.
        ttl: the lock time to live (TTL) in seconds. It should be higher than the expected run time of the cookbook.

    """

    suffix: str
    concurrency: int
    ttl: int


class CookbookBase(metaclass=ABCMeta):
    """Base Cookbook class that all cookbooks must extend to use the class-based API to define a cookbook.

    The class will be instantiated without any parameter.
    """

    # --- Reserved for Spicerack internal usage
    spicerack_path: str
    """Reserved class property used by Spicerack internally to track the Cookbook's path."""
    spicerack_name: str
    """Reserved class property used by Spicerack internally to track the Cookbook's name."""
    # ---

    owner_team: str = "unowned"
    """Name of the team owning this cookbook and responsible to keep it up to date. If unset and any parent package
    (directory of cookbooks) has the ``__owner_team__`` property set it will inherit it. It shows up when listing
    cookbooks and in the help message as parser epilog."""
    argument_task_required: Optional[bool] = None
    """Control if a ``-t/--task-id`` argument is included in the default argument parser for the Phabricator task ID.

        * If set to :py:data:`True` it will add a ``-t/--task-id`` required argument, accesible as ``args.task_id``.
        * If set to :py:data:`False` it will add a ``-t/--task-id`` optional argument, accessible as ``args.task_id``.
        * If set to :py:data:`None` it will not add the argument for providing a task ID.
        * When adding the argument it also validates that it is a valid Phabricator task ID (e.g. T12345).
    """
    argument_reason_required: Optional[bool] = None
    """Control if a ``-r/--reason`` argument is included in the default argument parser for the administrative reason.

        * If set to :py:data:`True` it will add a ``-r/--reason`` required argument, accesible as ``args.reason``.
        * If set to :py:data:`False` it will add a ``-r/--reason`` optional argument, accessible as ``args.reason``.
        * If set to :py:data:`None` it will not add the argument for providing ad administrative reason.
        * When adding the argument it also validates that it is a valid administrative reason (e.g. doesn't contains
          double quotes and is not empty).
    """

    def __init__(self, spicerack: Spicerack):
        """Initialize the instance and store the Spicerack instance into ``self.spicerack``.

        Arguments:
            spicerack: the Spicerack accessor instance with which the cookbook can access all the Spicerack
                capabilities.

        """
        self.spicerack = spicerack

    @property
    def title(self) -> str:
        """Returns the title of the Cookbook, must be a single line string.

        The default implementation returns the first line of the class docstring if there is any, a single
        dash otherwise.
        """
        if self.__doc__ is None:
            return "-"

        return self.__doc__.splitlines()[0]

    def argument_parser(self) -> argparse.ArgumentParser:
        """Optionally define an argument parser for the cookbook, if the cookbook accepts CLI arguments.

        The default implementation returns an empty ``ArgumentParser`` instance that doesn't accept any arguments and
        uses the class docstring as description. Based on the class parameters ``argument_*``, additional common
        arguments can be added automatically to the default argument parser.
        The actual command line arguments will be parsed by the Spicerack framework.

        Returns:
            the argument parser instance.

        """
        parser = argparse.ArgumentParser(description=self.__doc__, formatter_class=ArgparseFormatter)

        if self.argument_task_required is not None:

            def task_id_type(task_id: str) -> str:
                """Validates that the provided task ID matches a Phabricator task ID format.

                Arguments:
                    task_id: the task ID provided as argument.

                Returns:
                    the task ID if valid.

                Raises;
                    argparse.ArgumentTypeError: if the task ID is not valid.

                """
                pattern = r"T\d{1,6}$"
                if re.match(pattern, task_id) is None:
                    raise argparse.ArgumentTypeError(
                        f"Invalid Phabricator task ID, expected to match pattern '{pattern}', got '{task_id}'."
                    )

                return task_id

            parser.add_argument(
                "-t",
                "--task-id",
                required=self.argument_task_required,
                type=task_id_type,
                help="The Phabricator task ID (e.g. T12345).",
            )

        if self.argument_reason_required is not None:

            def reason_type(reason: str) -> str:
                """Validates that the provided administrative reason is valid.

                Arguments:
                    reason: the administrative reason.

                Returns:
                    the administrative reason if valid.

                Raises;
                    argparse.ArgumentTypeError: if the administrative reason is not valid.

                """
                if not reason:
                    raise argparse.ArgumentTypeError("The administrative reason cannot be empty.")

                if '"' in reason:
                    raise argparse.ArgumentTypeError(
                        f"The administrative reason cannot contain double quotes, got '{reason}'."
                    )

                return reason

            parser.add_argument(
                "-r",
                "--reason",
                required=self.argument_reason_required,
                type=reason_type,
                help="Administrative Reason. The current username and originating host are automatically added.",
            )

        return parser

    @abstractmethod
    def get_runner(self, args: argparse.Namespace) -> "CookbookRunnerBase":
        """Return the runner object that will be used to execute the cookbook.

        Derived classes must override this method and can perform any initialization and validation of the parsed
        arguments, but must not perform any real action here.

        If any exception is raised in this method Spicerack will not execute the cookbook.

        Arguments:
            args: the parsed arguments that were parsed using the defined ``argument_parser()``.

        Raises:
            BaseException: any exception raised in the ``get_runner()`` method will be catched by Spicerack and the
                Cookbook will not be executed.

        Returns:
            Must return an instance of a custom class derived from :py:class:`spicerack.cookbook.CookbookRunnerBase`
            that implements the actual execution of the cookbook.

        """


class CookbookRunnerBase(metaclass=ABCMeta):
    """Base class that all cookbooks must extend to use the class-based API to define the execution plan.

    The constructor of the class is left outside of the interface contract so that each cookbook is free to customize
    it as needed.
    """

    max_concurrency: int = 20
    """How many parallel runs of a specific cookbook inheriting from this class are accepted. If the ``lock_args``
    property is defined this one is ignored."""
    lock_ttl: int = 1800
    """The concurrency lock time to live (TTL) in seconds. For each concurrent run a lock is acquired for this amount
    of seconds. If the ``lock_args`` property is defined this one is ignored."""

    @property
    def runtime_description(self) -> str:
        """Optional message to be used as the runtime description of the cookbook.

        Cookbooks can override this instance property to define their custom description, also based on the given
        command line arguments. For example this will be used in the task start/end messages. The default
        implementation returns an empty string.
        """
        return ""

    @property
    def lock_args(self) -> LockArgs:
        """Optional property to dynamically modify the arguments used for the distributed lock of the cookbook runs.

        It is useful to allow to set a different concurrency and TTL for the cookbook's lock  based on the operation
        performed. For example allowing for an increased concurrency when performing read-only operations, or to
        differentiate the lock based on the given arguments, say having different locks for different datacenters or
        clusters.
        This property by default returns the static ``max_concurrency`` and ``lock_ttl`` class properties with an
        empty suffix.

        Returns:
            the arguments to be used by Spicerack to acquire the lock for the current cookbook.

        """
        return LockArgs(suffix="", concurrency=self.max_concurrency, ttl=self.lock_ttl)

    @property
    def skip_start_sal(self) -> bool:
        """Dynamically skip the START log to SAL. For fast cookbooks where it's ok to log just their completion.

        Returns:
            If set to :py:data:`True` Spicerack will skip logging the START of the cookbook to SAL and log to SAL only
            the end of the cookbook run with the keyword ``DONE`` instead of the usual ``END``.

        """
        return False

    @abstractmethod
    def run(self) -> Optional[int]:
        """Execute the cookbook.

        Returns:
            The return code of the cookbook, it should be zero or :py:data:`None` on success, a positive
            integer smaller than ``128`` and not in the range ``90-99``
            (see :ref:`Reserved exit codes<reserved-codes>`) in case of failure.

        Raises:
            BaseException: any exception raised in the ``run()`` method will be catched by Spicerack and the Cookbook
                execution will be considered failed.

        """

    def rollback(self) -> None:
        """Called by Spicerack when the cookbook fails the execution.

        This method by default does nothing. Cookbooks classes that inherit from this one can override it to add their
        own custom actions to perform on error to rollback to a previous state.
        The method will be called if the cookbook raises any un-caught exception or exits with a non-zero exit code.
        For example it can be used to cleanup any left-over inconsistent state as if the cookbook was never run.

        Any un-caught exception raised by this method will be caught by Spicerack and logged, along with the original
        exit code of the cookbook. The final exit code will be the reserved
        :py:const:`cookbooks.ROLLBACK_FAIL_RETCODE`.
        """



"""Log module."""

import logging
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import Optional

from wmflib.interactive import notify_logger
from wmflib.irc import SALSocketHandler, SocketHandler

root_logger = logging.getLogger()
irc_logger = logging.getLogger("spicerack_irc_announce")
sal_logger = logging.getLogger("spicerack_sal_announce")


class FilterOutCumin(logging.Filter):
    """A logging output filter to filter out Cumin's logs."""

    def filter(self, record: logging.LogRecord) -> bool:
        """Filter out Cumin's log messages.

        Arguments:
            record: the logging record.

        Returns:
            :py:data:`False` if the record should be filtered out, :py:data:`True` if it should be included.
            According to Python's logging interface, see: https://docs.python.org/3/library/logging.html#filter-objects

        """
        if record.name == "cumin" or record.name.startswith("cumin."):
            return False  # Filter it out

        return True


def setup_logging(
    base_path: Path,
    name: str,
    user: str,
    *,
    dry_run: bool = True,
    host: Optional[str] = None,
    port: int = 0,
    notify_logger_enabled: bool = False,
) -> None:
    """Setup the root logger instance.

    Arguments:
        base_path: the base path where to save the logs.
        name: the name of log file to use without extension.
        user: the username for the IRC logging.
        dry_run: whether this is a dry-run.
        host: the tcpircbot hostname for the IRC logging.
        port: the tcpircbot port for the IRC logging.
        notify_logger_enabled: whether to setup wmflib's notify_logger notification to IRC.

    """
    base_path.mkdir(mode=0o755, parents=True, exist_ok=True)

    if dry_run:
        dry_run_prefix = "DRY-RUN "
    else:
        dry_run_prefix = ""

    # Default INFO logging
    formatter = logging.Formatter(fmt=f"%(asctime)s {dry_run_prefix}{user} %(process)d [%(levelname)s] %(message)s")
    # Tentatively keep logs forever for auditing purposes. Limit them to 10MB each file and keep 500 files.
    # Max theoretical space used for the standard logs per cookbook is ~5GB
    handler = RotatingFileHandler(base_path / f"{name}.log", maxBytes=(10 * (1024**2)), backupCount=500)
    handler.setFormatter(formatter)
    handler.setLevel(logging.INFO)

    # Extended logging for detailed debugging
    formatter_extended = logging.Formatter(
        fmt=(
            f"%(asctime)s {dry_run_prefix}{user} %(process)d [%(levelname)s %(name)s:%(lineno)s in %(funcName)s] "
            "%(message)s"
        )
    )
    # Tentatively keep logs forever for auditing purposes. Limit them to 10MB each file and keep 500 files.
    # Max theoretical space used for the extended logs per cookbook is ~5GB
    handler_extended = RotatingFileHandler(
        base_path / f"{name}-extended.log", maxBytes=(10 * (1024**2)), backupCount=500
    )
    handler_extended.setFormatter(formatter_extended)
    handler_extended.setLevel(logging.DEBUG)

    # Stderr logging
    output_handler = logging.StreamHandler()
    if dry_run:
        output_handler.setFormatter(logging.Formatter(fmt="DRY-RUN: %(message)s"))
        output_handler.setLevel(logging.DEBUG)
    else:
        output_handler.setLevel(logging.INFO)
    output_handler.addFilter(FilterOutCumin())

    root_logger.addHandler(handler)
    root_logger.addHandler(handler_extended)
    root_logger.addHandler(output_handler)
    root_logger.setLevel(logging.DEBUG)

    if not dry_run and host is not None and port > 0:
        irc_logger.addHandler(SocketHandler(host, port, user))
        irc_logger.setLevel(logging.INFO)
        sal_logger.addHandler(SALSocketHandler(host, port, user))
        sal_logger.setLevel(logging.INFO)

        if notify_logger_enabled:
            notify_handler = SocketHandler(host, port, user)
            notify_formatter = logging.Formatter(fmt=f"{name} (PID %(process)d) %(message)s")
            notify_handler.setFormatter(notify_formatter)
            notify_logger.addHandler(notify_handler)
            notify_logger.setLevel(logging.INFO)

    # Silence external noisy loggers
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)

    # Elasticsearch lib is very noisy about HTTP level errors
    # ideally, we'd want to keep it at WARNING level for logs
    # sent to file, but ERROR for the console. Since this is
    # non trivial, let's raise level to ERROR for the moment.
    logging.getLogger("elasticsearch").setLevel(logging.ERROR)


def log_task_start(*, skip_start_sal: bool, message: str) -> None:
    """Log the start of a task on the logs and unless ``skip_start_sal`` is :py:data:`True` also to IRC/SAL.

    Arguments:
        skip_start_sal: whether to skip the IRC/SAL logging for the task start.
        message: the message to be logged.

    """
    if skip_start_sal:
        root_logger.info("START - %s", message)
    else:
        sal_logger.info("START - %s", message)


def log_task_end(*, skip_start_sal: bool, status: str, message: str) -> None:
    """Log the end of a task both on the logs and IRC.

    Arguments:
        status: the final status of the task.
        message: the message to be logged.
        skip_start_sal: whether this is the only log for the task or a regular ``START``/``END`` one.

    """
    prefix = "DONE" if skip_start_sal else "END"
    sal_logger.info("%s (%s) - %s", prefix, status, message)



"""Hosts module tests."""

import re
from unittest import mock

import pytest

from spicerack import Spicerack, hosts
from spicerack.alerting import AlertingHosts
from spicerack.alertmanager import AlertmanagerHosts
from spicerack.apt import AptGetHosts
from spicerack.exceptions import SpicerackError
from spicerack.icinga import IcingaHosts
from spicerack.ipmi import Ipmi
from spicerack.mysql import MysqlRemoteHosts
from spicerack.netbox import NetboxServer
from spicerack.puppet import PuppetHosts
from spicerack.redfish import Redfish
from spicerack.remote import RemoteHosts
from spicerack.tests import SPICERACK_TEST_PARAMS


class TestHost:
    """Test class for the Host class."""

    def _setup(self, name, monkeypatch, netbox_host):
        """Initiliaze the fixtures."""
        monkeypatch.setenv("SUDO_USER", "user1")
        monkeypatch.setenv("MGMT_PASSWORD", "env_password")
        proxy = "http://proxy.example.com:8080"
        self.mocked_pynetbox.return_value.dcim.devices.get.return_value = netbox_host
        mgmt_address = mock.MagicMock()
        mgmt_address.address = "10.0.0.1/16"
        mgmt_address.dns_name = f"{name}.mgmt.example.com"
        self.mocked_pynetbox.return_value.ipam.ip_addresses.get.return_value = mgmt_address
        return Spicerack(verbose=True, dry_run=False, http_proxy=proxy, **SPICERACK_TEST_PARAMS)

    def setup_method(self):
        """Initialize the test for each method."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_pynetbox = mock.patch("pynetbox.api").start()

    def teardown_method(self):
        """Cleanup any leftover patching."""
        self.mocked_pynetbox.stop()

    def test_host_from_remote_ok(self, monkeypatch, netbox_host):
        """It should return a Host instance from a RemoteHosts instance with just one host."""
        spicerack = self._setup("physical", monkeypatch, netbox_host)
        remote = spicerack.remote().query("D{physical.example.com}")
        host = hosts.Host.from_remote(remote, spicerack)
        assert isinstance(host, hosts.Host)
        assert host.hostname == "physical"

    def test_host_from_remote_fail(self, monkeypatch, netbox_host):
        """It should raise a HostError exception if the remote instance has multiple hosts."""
        spicerack = self._setup("physical", monkeypatch, netbox_host)
        remote = spicerack.remote().query("D{host[1-5].example.com}")
        with pytest.raises(
            hosts.HostError,
            match=re.escape(
                "Unable to create Host instance from RemoteHosts host[1-5].example.com, expected 1 host, got 5."
            ),
        ):
            hosts.Host.from_remote(remote, spicerack)

    def test_accessors_physical(self, monkeypatch, netbox_host):
        """An Host instance for a physical server should allow to access library features for a specific host."""
        name = "physical"
        spicerack = self._setup(name, monkeypatch, netbox_host)
        host = hosts.Host(name, spicerack)
        assert host.hostname == name
        assert host.fqdn == f"{name}.example.com"
        assert host.mgmt_fqdn == f"{name}.mgmt.example.com"
        assert isinstance(host.remote(), RemoteHosts)
        assert str(host.remote()) == f"{name}.example.com"
        assert isinstance(host.netbox(), NetboxServer)
        assert isinstance(host.puppet(), PuppetHosts)
        assert isinstance(host.mysql(), MysqlRemoteHosts)
        assert isinstance(host.apt_get(), AptGetHosts)
        assert isinstance(host.alertmanager(), AlertmanagerHosts)
        assert isinstance(host.redfish(), Redfish)
        assert isinstance(host.ipmi(), Ipmi)

    def test_accessors_virtual(self, monkeypatch, netbox_virtual_machine):
        """An Host instance for a virtual machine should allow to access library features for a specific host."""
        name = "virtual"
        spicerack = self._setup(name, monkeypatch, netbox_virtual_machine)
        host = hosts.Host(name, spicerack)
        assert host.hostname == name
        assert host.fqdn == f"{name}.example.com"
        assert isinstance(host.remote(), RemoteHosts)
        assert str(host.remote()) == f"{name}.example.com"
        assert isinstance(host.netbox(), NetboxServer)
        assert isinstance(host.puppet(), PuppetHosts)
        assert isinstance(host.mysql(), MysqlRemoteHosts)
        assert isinstance(host.apt_get(), AptGetHosts)
        assert isinstance(host.alertmanager(), AlertmanagerHosts)
        # No mgmt_fqdn, ipmi or redfish, they should raise an exception

    @mock.patch("spicerack.Dns", autospec=True)
    @mock.patch("spicerack.remote.Remote.query", autospec=True)
    @mock.patch("spicerack.icinga.CommandFile", autospec=True)
    def test_alerting_icinga(self, mocked_command_file, mocked_remote_query, mocked_dns, monkeypatch, netbox_host):
        """When accessing the alerting or icinga accessors should return an instance of the respecive classes."""
        spicerack = self._setup("physical", monkeypatch, netbox_host)
        host = hosts.Host("physical", spicerack)
        mocked_command_file.return_value = "/var/lib/icinga/rw/icinga.cmd"
        icinga_server = mock.MagicMock(spec_set=RemoteHosts)
        icinga_server.hosts = "icinga-server.example.com"
        icinga_server.__len__.return_value = 1
        mocked_remote_query.return_value = icinga_server
        mocked_dns.return_value.resolve_cname.return_value = "icinga-server.example.com"
        assert isinstance(host.alerting(), AlertingHosts)
        assert isinstance(host.icinga(), IcingaHosts)

    def test_init_not_found(self, monkeypatch, netbox_host):
        """It should raise HostError if unable to find the host in Netbox."""
        spicerack = self._setup("nonexistent", monkeypatch, netbox_host)
        self.mocked_pynetbox.return_value.dcim.devices.get.return_value = None
        self.mocked_pynetbox.return_value.virtualization.virtual_machines.get.return_value = None

        with pytest.raises(hosts.HostError, match="Unable to find host nonexistent in Netbox"):
            hosts.Host("nonexistent", spicerack)

    @pytest.mark.parametrize(
        "attribute, error",
        (
            ("mgmt_fqdn", "Server virtual is a virtual machine, does not have a management address."),
            ("redfish", "Host virtual is not a Physical server, Redfish is not supported."),
            ("ipmi", "Host 'virtual' is a Virtual Machine, IPMI not supported."),
        ),
    )
    def test_ipmi_redfish_virtual(self, attribute, error, monkeypatch, netbox_virtual_machine):
        """For virtual machines accessing ipmi or redfish should raise an exception."""
        spicerack = self._setup("virtual", monkeypatch, netbox_virtual_machine)
        host = hosts.Host("virtual", spicerack)

        with pytest.raises(SpicerackError, match=re.escape(error)):
            attr = getattr(host, attribute)
            if callable(attr):
                attr()



"""dbctl module."""

from typing import Any

from conftool.extensions.dbconfig.config import DbConfig
from conftool.extensions.dbconfig.entities import Instance, Section

from spicerack.confctl import Confctl


class Dbctl(Confctl):
    """Extend the Confctl class to support dbctl actions."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the instance.

        Parameters as required by the parent class.

        """
        super().__init__(*args, **kwargs)
        schema = self._client.schema
        self._mediawiki_config = DbConfig(schema, Instance(schema), Section(schema), self._client.configuration.dbctl())
        self._dbconfig_instance = Instance(schema, self._mediawiki_config.check_instance)
        self._dbconfig_section = Section(schema, self._mediawiki_config.check_section)

    @property
    def config(self) -> DbConfig:
        """Getter for the instance to act like `dbctl config`."""
        return self._mediawiki_config

    @property
    def instance(self) -> Instance:
        """Getter for the instance to act like `dbctl instance`."""
        return self._dbconfig_instance

    @property
    def section(self) -> Section:
        """Getter for the instance to act like `dbctl section`."""
        return self._dbconfig_section



"""Group3 get argument_parser() raise SystemExit(str)."""


def argument_parser():
    """As required by spicerack._cookbook."""
    raise SystemExit("argument_parser")


def run(_args, _spicerack):
    """As required by spicerack._cookbook."""



"""Hosts module."""

from typing import TYPE_CHECKING

from spicerack.exceptions import SpicerackError
from spicerack.mysql import MysqlRemoteHosts
from spicerack.netbox import NetboxHostNotFoundError
from spicerack.remote import RemoteHosts

if TYPE_CHECKING:  # Prevent circular dependency, is needed only for type hints
    import spicerack  # pragma: no cover | not imported at runtime


class HostError(SpicerackError):
    """Custom exception class for errors related to the Host instance."""


class Host:
    """A class to represent a host across various services.

    The class ensures that the host exists in our source of truth (Netbox) and exposes various services that will have
    the host as target for easy of use when managing a single host.
    """

    def __init__(
        self, name: str, spicerack_instance: "spicerack.Spicerack", *, netbox_read_write: bool = False
    ) -> None:
        """Initialize the instance.

        Arguments:
            name: the short hostname of the host.
            spicerack_instance: the spicerack instance.
            netbox_read_write: whether the Netbox related operation should be performed with a read-write token
                (:py:data:`True`) or a read-only one (:py:data:`False`).

        Raises:
            spicerack.netbox.NetboxError: if unable to find the host in Netbox or load its data.

        """
        self._spicerack = spicerack_instance
        self._remote = self._spicerack.remote()
        try:
            self._netbox_server = self._spicerack.netbox_server(name, read_write=netbox_read_write)
        except NetboxHostNotFoundError:
            raise HostError(f"Unable to find host {name} in Netbox") from None

    @classmethod
    def from_remote(cls, remote_hosts: RemoteHosts, spicerack_instance: "spicerack.Spicerack") -> "Host":
        """Initialize the Host instance from a RemoteHosts instance.

        Arguments:
            remote_hosts: the intance from where to create the host instance, must have only one host.
            spicerack_instance: the spicerack instance to pass to the Host constructor.

        Returns:
            the Host instance.

        Raises:
            spicerack.hosts.HostError: if the remote hosts doesn't have a single host.

        """
        if len(remote_hosts) != 1:
            raise HostError(
                f"Unable to create Host instance from RemoteHosts {remote_hosts}, "
                f"expected 1 host, got {len(remote_hosts)}."
            )

        return cls(str(remote_hosts).split(".", maxsplit=1)[0], spicerack_instance)

    @property
    def hostname(self) -> str:
        """The short hostname of the host.

        Examples:
            ::

                >>> host.hostname
                'example1001'

        Returns:
            the hostname as reported on Netbox.

        """
        return self._netbox_server.name

    @property
    def fqdn(self) -> str:
        """The fully qualified domain name (FQDN) of the host.

        Examples:
            ::

                >>> host.fqdn
                'example1001.eqiad.wmnet'

        Returns:
            the FQDN as defined in Netbox.

        """
        return self._netbox_server.fqdn

    @property
    def mgmt_fqdn(self) -> str:
        """The fully qualified domain name (FQDN) of the management interface of the host.

        Examples:
            ::

                >>> host.mgmt_fqdn
                'example1001.mgmt.eqiad.wmnet'


        Returns:
            the FQDN of the management interface as defined in Netbox.

        """
        return self._netbox_server.mgmt_fqdn

    def remote(self) -> "spicerack.RemoteHosts":
        """Get an instance to execute ssh commands on the host. It ensures that the host is present in PuppetDB.

        Examples:
            ::

                >>> host.remote.run_sync('command')

        Returns:
            the remote hosts instance.

        """
        return self._remote.query(self.fqdn)

    def netbox(self) -> "spicerack.NetboxServer":
        """Get an instance with all the Netbox data of the host.

        Examples:
            ::

                >>> host.netbox.status
                'active'

        Returns:
            the netbox server instance.

        """
        return self._netbox_server

    def puppet(self) -> "spicerack.PuppetHosts":
        """Get an instance to manage Puppet on the host.

        Examples:
            ::

                >>> host.puppet.run()

        Returns:
            the Puppet hosts instance.

        """
        return self._spicerack.puppet(self.remote())

    def mysql(self) -> MysqlRemoteHosts:
        """Get an instance to manage Mysql/Mariadb on the host.

        There is no check that the host has a Mysql/Mariadb server when calling this property.
        The Mysql/Mariadb specific features will fail if the host doesn't have a server installed/running.

        Examples:
            ::

                >>> host.mysql.run_query(query)

        Returns:
            the mysql remote hosts instance.

        """
        return MysqlRemoteHosts(self.remote())

    def apt_get(self) -> "spicerack.AptGetHosts":
        """Get an instance to manage Debian packages on the host via apt-get.

        Examples:
            ::

                >>> host.apt_get.update()

        Returns:
            the apt-get instance.

        """
        return self._spicerack.apt_get(self.remote())

    def alerting(self) -> "spicerack.AlertingHosts":
        """Get an instance to manage both Alertmanager and Icinga alerts for this host.

        Examples:
            ::

                >>> with host.alerting.downtimed(reason, duration=duration):
                ...     # do something

        Returns:
            the alerting hosts instance.

        """
        return self._spicerack.alerting_hosts([self.fqdn])

    def icinga(self) -> "spicerack.IcingaHosts":
        """Get an instance to manage Icinga alerts for this host.

        Examples:
            ::

                >>> with host.icinga.downtimed(reason, duration=duration):
                ...     # do something

        Returns:
            the Icinga hosts instance.

        """
        return self._spicerack.icinga_hosts([self.fqdn])

    def alertmanager(self) -> "spicerack.AlertmanagerHosts":
        """Get an instance to manage Alertmanager alerts for this host.

        Examples:
            ::

                >>> with host.alertmanager.downtimed(reason, duration=duration):
                ...     # do something

        Returns:
            the Alertmanager hosts instance.

        """
        return self._spicerack.alertmanager_hosts([self.fqdn])

    def redfish(self) -> "spicerack.Redfish":
        """Get an instance to manage the host using Redfish on the management interface of the host.

        Examples:
            ::

                >>> host.redfish.check_connection()

        Returns:
            the Redfish instance.

        Raises:
            spicerack.SpicerackError: if the host is a Virtual Machine or the manufacturer is not supported.

        """
        return self._spicerack.redfish(self.hostname)

    def ipmi(self) -> "spicerack.Ipmi":
        """Get an instance to manage the host using IPMI on the management interface of the host.

        See Also:
            https://wikitech.wikimedia.org/wiki/Management_Interfaces

        Examples:
            ::

                >>> host.ipmi.power_status()
                'on'

        Returns:
            the ipmi instance.

        Raises:
            spicerack.hosts.HostError: if the host is a Virtual Machine.

        """
        if self.netbox().virtual:
            raise HostError(f"Host '{self.hostname}' is a Virtual Machine, IPMI not supported.")

        return self._spicerack.ipmi(self.mgmt_fqdn)



"""Pytest shared fixtures."""

import json
from types import SimpleNamespace
from unittest import mock

import pytest


class NetboxObject(SimpleNamespace):
    """Simple object to represent a pynetbox API response with a save() method and dict representation."""

    def __iter__(self):
        """Make the object convertable to dict."""
        # The JSON passage is needed to recursively convert all the NetboxObject instances to dict
        return iter(
            json.loads(json.dumps(self, default=lambda x: {i: j for i, j in x.__dict__.items() if i != "save"})).items()
        )


def _base_netbox_obj(name, additional_properties):
    """Return a simple object to represent a response from Netbox API."""
    dict_obj = {
        "name": name,
        "id": 1,
        "asset_tag": "ASSET1234",
        "status": {"value": "active", "label": "Active"},
        "primary_ip4": {
            "id": 1,
            "family": 4,
            "address": "10.0.0.1/22",
            "dns_name": f"{name}.example.com",
        },
        "primary_ip6": {
            "id": 1,
            "family": 6,
            "address": "2620:0:861:103:10::1/64",
            "dns_name": f"{name}.example.com",
        },
        "role": {
            "id": 1,
            "name": "Server",
            "slug": "server",
        },
        "device_type": {
            "manufacturer": {
                "slug": "dell",
            },
        },
    }
    dict_obj["primary_ip"] = dict_obj["primary_ip6"]
    dict_obj["primary_ip"]["assigned_object"] = {
        "id": 1,
        "connected_endpoints": [{"untagged_vlan": {"name": "test_vlan"}}],
        "type": {"value": "10gbase-x-sfpp"},
    }
    dict_obj.update(additional_properties)

    def custom_hook(decoded_dict):
        """Custom hook for JSON load to convert a dict to an object with a save() attribute."""
        decoded_obj = NetboxObject(**decoded_dict)
        decoded_obj.save = mock.MagicMock(return_value=True)  # pylint: disable=attribute-defined-outside-init
        return decoded_obj

    obj = json.loads(json.dumps(dict_obj), object_hook=custom_hook)
    obj.status.__str__ = lambda: dict_obj["status"]["label"]

    return obj


@pytest.fixture()
def netbox_host():
    """Return a mocked Netbox physical device."""
    return _base_netbox_obj("physical", {"rack": {"id": 1, "name": "rack1"}, "cluster": None})


@pytest.fixture()
def netbox_virtual_machine():
    """Return a mocked Netbox virtual machine."""
    return _base_netbox_obj("virtual", {"cluster": {"id": 1, "name": "testcluster"}})



"""Generic API client module."""

import logging
from typing import Any

from requests import Response, Session
from requests.exceptions import RequestException

from spicerack.exceptions import SpicerackError

logger = logging.getLogger(__name__)


class APIClientError(SpicerackError):
    """General errors raised by this module."""


class APIClientResponseError(APIClientError):
    """Exception class for failed responses to API requests."""

    def __init__(self, response: Response, message: str = "") -> None:
        """Override parent constructor to add the requests's response object.

        Arguments:
            response: the requests's response object.
            message: an optional exception message. A default message with the HTTP method, URL and status code will
                be prefixed before this message.

        """
        method = str(response.request.method).upper()
        super().__init__(f"{method} {response.request.url} returned HTTP {response.status_code} {message}".strip())
        self.response = response


class APIClient:
    """Generic API client class with DRY-RUN support. API-specific classes should derive from this one."""

    def __init__(self, base_url: str, http_session: Session, *, dry_run: bool = True) -> None:
        """Initialize the instance.

        Arguments:
            base_url: the full base URL for the API. It must include the scheme and the domain and can include the
                API path prefix if there is one.
            http_session: the requests's session to use to make the API calls.
            dry_run: whether this is a DRY-RUN.

        """
        self._base_url = base_url
        self._http_session = http_session
        self._dry_run = dry_run

    @property
    def http_session(self) -> Session:
        """Getter for the HTTP session request object to adjust it's configuration if needed.

        Returns:
            the pre-configured session.

        """
        return self._http_session

    def request(self, method: str, uri: str, **kwargs: Any) -> Response:
        """Perform a request against the HTTP session with the provided HTTP method and data.

        Arguments:
            method: the HTTP method to use (e.g. "get").
            uri: the relative URI to request.
            **kwargs: arbitrary keyword arguments, to be passed to the requests library.

        Raises:
            spicerack.apiclient.APIClientError: if the given uri does not start with a slash (/) or the request
                couldn't be performed.
            spicerack.apiclient.APIClientResponseError: if the response status code is not ok, between 400 and 600.

        Returns:
            The API response object if not in DRY-RUN mode or the request is read-only (GET/HEAD/OPTIONS). When in
            DRY-RUN mode and a read-write request is made the API will not be called and a dummy HTTP 200 response
            will be returned.

        """
        if uri[0] != "/":
            raise APIClientError(f"Invalid uri '{uri}', it must start with a /")

        url = f"{self._base_url}{uri}"

        if self._dry_run and method.lower() not in ("head", "get", "options"):  # RW call
            logger.info("Would have called %s on %s", method, url)
            return self._get_dummy_response()

        try:
            response = self._http_session.request(method, url, **kwargs)
        except RequestException as e:
            message = f"Failed to perform {method.upper()} request to {url}"
            if self._dry_run:
                logger.error("%s: %s", message, e)
                return self._get_dummy_response()

            raise APIClientError(message) from e

        if not response.ok:
            raise APIClientResponseError(response)

        return response

    @staticmethod
    def _get_dummy_response() -> Response:
        """Return a dummy requests's Response to be used in DRY-RUN mode."""
        response = Response()
        response.status_code = 200
        return response
